import datetime
import importlib
import os
from enum import Enum
from typing import Any, Callable, Dict, Iterator, Optional, Type, cast

from dateutil.parser import parse as parse_date

from redun.hashing import hash_bytes, hash_struct, hash_tag_bytes
from redun.utils import get_func_source, iter_nested_value, pickle_dumps, pickle_loads

MIME_TYPE_PICKLE = "application/python-pickle"


class TypeError(Exception):
    pass


class InvalidValueError(Exception):
    pass


# Python does not provide a type for None, so we compute it here.
NoneType: Type = type(None)


def get_raw_type_name(raw_type: Any) -> str:
    return "{}.{}".format(raw_type.__module__, raw_type.__name__)


class TypeRegistry:
    """
    A registry for redun types and a dispatch for value methods.
    """

    def __init__(self):
        self._type_name2value_type: Dict[str, Type["Value"]] = {}
        self._raw2proxy_type: Dict[Any, Type["ProxyValue"]] = {}

    def get_type_name(self, raw_type: Type) -> str:
        """
        Returns the type name for a python type.
        """
        value_type = self.get_type(raw_type, use_default=False)
        if value_type:
            assert value_type.type_name
            return value_type.type_name

        # All Value classes should be registered at this point.
        assert not issubclass(raw_type, Value)

        return get_raw_type_name(raw_type)

    def parse_type_name(self, type_name: str) -> Type:
        """
        Parse a type name into a python type.
        """
        if type_name == "builtins.NoneType":
            return NoneType

        value_type = self._type_name2value_type.get(type_name)
        if value_type:
            return value_type

        try:
            module_name, raw_type_name = type_name.rsplit(".", 1)
            return importlib.import_module(module_name).__dict__[raw_type_name]
        except (ValueError, ModuleNotFoundError, KeyError):
            raise TypeError('Unable to import type "{}"'.format(type_name))

    def register(self, value_type: Type["Value"]) -> None:
        """
        Register a Value class with the type system.
        """
        if getattr(value_type, "__args__", None):
            # In python 3.6, parameterized types try to register themselves
            # and can conflict with registering the unparameterized type.
            # We do not need them registered, so we ignore them here.
            return

        # Note: getattr is used because Values start registering even before ProxyValue is defined.
        if getattr(value_type, "proxy", False) and getattr(value_type, "type", None):
            # value_type is a proxy for another raw type.
            self._raw2proxy_type[getattr(value_type, "type")] = cast(Type[ProxyValue], value_type)

        assert isinstance(value_type.type_name, str)
        self._type_name2value_type[cast(str, value_type.type_name)] = value_type

    def _get_proxy_type(self, raw_type: type) -> Optional[Type["Value"]]:
        """
        Search through a type's superclasses for a registered ProxyValue.
        """
        for super_raw_type in raw_type.__class__.mro(raw_type):
            proxy_type = self._raw2proxy_type.get(super_raw_type)
            if proxy_type:
                self._raw2proxy_type[raw_type] = proxy_type
                return proxy_type
        return None

    def get_value(self, raw_value: Any) -> "Value":
        """
        Return a Value instance for a raw python type.
        """
        raw_type = type(raw_value)
        proxy_type = self._get_proxy_type(raw_type)
        if proxy_type:
            return proxy_type(raw_value)

        elif isinstance(raw_value, Value):
            return raw_value

        else:
            return ProxyValue(raw_value)

    def get_type(self, raw_type: Type, use_default: bool = True) -> Optional[Type["Value"]]:
        """
        Return a registered Value class for a raw python type.
        """
        proxy_type = self._get_proxy_type(raw_type)
        if proxy_type:
            return proxy_type

        elif isinstance(raw_type, type) and issubclass(raw_type, Value):
            return raw_type

        elif use_default:
            return ProxyValue

        else:
            return None

    def is_valid(self, value: Any) -> bool:
        """
        Returns True if the value is still good to use since being cached.

        For example, if the value was a File reference and the corresponding
        file on the filesystem has since been removed/altered, the File
        reference is no longer valid.
        """
        return self.get_value(value).is_valid()

    def is_valid_nested(self, nested_value: Any) -> bool:
        """
        Returns True if nested value is still good to use since being cached.
        """
        return all(map(self.is_valid, iter_nested_value(nested_value)))

    def get_hash(self, value: Any, data: Optional[bytes] = None) -> str:
        """
        Hashes a value according to the hash method precedence.
        """
        return self.get_value(value).get_hash()

    def get_serialization_format(self, value: Any) -> str:
        """
        Returns mimetype of serialization.
        """
        return self.get_value(value).get_serialization_format()

    def serialize(self, value: Any) -> bytes:
        """
        Serializes the Value into a byte sequence.
        """
        return self.get_value(value).serialize()

    def deserialize(self, type_name: str, data: bytes) -> Any:
        """
        Returns a deserialization of bytes `data` into a new Value.
        """
        try:
            raw_type = self.parse_type_name(type_name)
        except TypeError:
            raise InvalidValueError(f"Unknown type: {type_name}")
        value_type = self.get_type(raw_type)
        assert value_type
        return value_type.deserialize(raw_type, data)

    def preprocess(self, raw_value: Any, preprocess_args: dict) -> Any:
        """
        Preprocess a value before passing to a Task.
        """
        return self.get_value(raw_value).preprocess(preprocess_args)

    def postprocess(self, raw_value: Any, postprocess_args: dict) -> Any:
        """
        Post process a value resulting from a Task.
        """
        return self.get_value(raw_value).postprocess(postprocess_args)

    def parse_arg(self, raw_type: Type, arg: str) -> Any:
        """
        Parse a command-line argument of type value_type.
        """
        value_type = self.get_type(raw_type, use_default=False)
        if value_type:
            return value_type.parse_arg(raw_type, arg)
        else:
            # Try to parse argument using raw_type constructor.
            return raw_type(arg)

    def iter_subvalues(self, raw_value: Any) -> Iterator["Value"]:
        """
        Iterates through the Value's subvalues.
        """
        return self.get_value(raw_value).iter_subvalues()


# Global singleton instance of TypeRegistry.
_type_registry = TypeRegistry()


def get_type_registry() -> TypeRegistry:
    """
    Return the global singleton instance of TypeRegistry.
    """
    return _type_registry


class MetaValue(type):
    """
    Registers Value classes as they are defined.
    """

    def __init__(cls, name, bases, dct, **kwargs):
        # Use cls.__dict__ to look for class attributes defined directly on class
        # and ignore what is set on super classes.
        if not cls.__dict__.get("type_name"):
            if cls.__dict__.get("type"):
                cls.type_name = get_raw_type_name(cls.type)
            else:
                cls.type_name = get_raw_type_name(cls)

        if getattr(cls, "register", True):
            registry = get_type_registry()
            registry.register(cls)


# In Python 3.6, in order to use metaclass for a Generic class we need to
# inherit from GenericMeta.
# https://github.com/python/typing/issues/449
try:
    # Types are ignored here because GenericMeta has been removed in python 3.7+
    from typing import GenericMeta  # type: ignore

    class MetaValue(MetaValue, GenericMeta):  # type: ignore
        def __init__(cls, name, bases, dct, **kwargs):
            super().__init__(name, bases, dct, **kwargs)

except ImportError:
    pass


class Value(metaclass=MetaValue):
    """
    Base class for input and output values in redun.
    """

    type_name: Optional[str] = None

    def __init__(self, *args, **kwargs):
        pass

    def is_valid(self) -> bool:
        """
        Returns True if the value is still good to use since being cached.

        For example, if the value was a File reference and the corresponding
        file on the filesystem has since been removed/altered, the File
        reference is no longer valid.
        """
        return True

    def get_hash(self, data: Optional[bytes] = None) -> str:
        """
        Returns a hash for the value.
        """
        if data is None:
            data = pickle_dumps(self)
        return hash_tag_bytes("Value", data)

    def get_serialization_format(self) -> str:
        """
        Returns mimetype of serialization.
        """
        return MIME_TYPE_PICKLE

    def serialize(self) -> bytes:
        """
        Serializes the Value into a byte sequence.
        """
        return pickle_dumps(self)

    @classmethod
    def deserialize(cls, raw_type: type, data: bytes) -> Any:
        """
        Returns a deserialization of bytes `data` into a new Value.
        """
        return pickle_loads(data)

    def __getstate__(self) -> dict:
        """
        Returns a plain python datastructure for serialization.
        """
        raise NotImplementedError()

    def __setstate__(self, state: dict) -> None:
        """
        Populates the value from state during deserialization.
        """
        raise NotImplementedError()

    def preprocess(self, preprocess_args) -> "Value":
        """
        Preprocess a value before passing to a Task.
        """
        return self

    def postprocess(self, postprocess_args) -> "Value":
        """
        Post process a value resulting from a Task.
        """
        return self

    @classmethod
    def parse_arg(cls, raw_type: type, arg: str) -> Any:
        """
        Parse a command line argument in a new Value.
        """
        # By default, try to use constructor to parse command line argument.
        return cls(arg)

    def iter_subvalues(self) -> Iterator["Value"]:
        """
        Iterates through the Value's subvalues.
        """
        if False:
            # By default, Values have no subvalues, but we need the yield
            # statement here to convert this method into an iterator.
            # Subclasses can then override this method for their own needs.
            yield
        return


class ProxyValue(Value):
    """
    Class for overriding Value behavior (hashing, serialization) for raw python types.
    """

    proxy = True
    type: Any = None

    def __init__(self, instance: Any):
        self.instance = instance

    def get_hash(self, data: Optional[bytes] = None) -> str:
        """
        Returns a hash for the value.
        """
        if data is None:
            data = pickle_dumps(self.instance)
        return hash_tag_bytes("Value", data)

    def serialize(self) -> bytes:
        """
        Serializes the Value into a byte sequence.
        """
        return pickle_dumps(self.instance)

    def preprocess(self, preprocess_args) -> "Value":
        """
        Preprocess a value before passing to a Task.
        """
        return self.instance

    def postprocess(self, postprocess_args) -> "Value":
        """
        Post process a value resulting from a Task.
        """
        return self.instance

    def iter_subvalues(self) -> Iterator["Value"]:
        """
        Iterates through the Value's subvalues.
        """
        if isinstance(self.instance, Value):
            # No subvalues.
            return
        for subvalue in iter_nested_value(self.instance):
            if isinstance(subvalue, Value):
                yield subvalue
                yield from subvalue.iter_subvalues()


class Bool(ProxyValue):
    """
    Augment builtins.bool to support argument parsing.
    """

    type = bool
    type_name = "builtins.bool"

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> Any:
        arg = arg.lower()
        if arg == "true":
            return True
        elif arg == "false":
            return False
        else:
            raise ValueError('Unknown bool "{}"'.format(arg))


class Set(ProxyValue):
    """
    Augment builtins.set to support stable hashing.
    """

    type = set
    type_name = "builtins.set"

    def get_hash(self, data: Optional[bytes] = None) -> str:
        # Sort the set to ensure stable serialization and hashing.
        bytes = pickle_dumps(sorted(self.instance))

        # Use a unique tag to distinguish from hashing a list.
        return hash_tag_bytes("Value.set", bytes)


class EnumType(ProxyValue):
    """
    Augment enum.Enum to support argument parsing with choices.
    """

    type = Enum
    type_name = "enum.Enum"

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> Any:
        name2item = {item.name: item for item in raw_type}
        try:
            if "." in arg:
                enum_name, arg = arg.split(".", 1)
                if enum_name != raw_type.__name__:
                    raise KeyError
            return name2item[arg]
        except KeyError:
            raise ValueError("{} is not a valid {}".format(arg, raw_type.__name__))


class DatetimeType(ProxyValue):
    """
    Augment datetime.datetime to support argument parsing.
    """

    type = datetime.datetime
    type_name = "datetime.datetime"

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> datetime.datetime:
        return parse_date(arg)


# Get the type for python functions.
function_type = type(lambda: None)


def make_unknown_function(func_name: str) -> Callable:
    """
    Returns a stub function in place of a function that could not be reimported.
    """

    def unknown_function(*args, **kwargs):
        raise ValueError(f"Function '{func_name}' cannot be found.")

    # Mark this function as an unknown function.
    unknown_function.unknown_function = True  # type: ignore
    return unknown_function


def is_unknown_function(func: Callable) -> bool:
    """
    Returns True if the function was unknown when trying to reimport.
    """
    return getattr(func, "unknown_function", False)


class Function(ProxyValue):
    """
    Value class to allow redun to hash and cache plain Python functions.
    """

    type = function_type
    type_name = "builtins.function"

    def __init__(self, instance: Callable):
        super().__init__(instance)
        self.module = instance.__module__
        self.name = instance.__name__
        self._hash = self._calc_hash()

        if not is_unknown_function(instance) and "<locals>" in instance.__qualname__:
            raise InvalidValueError(
                "functions used as redun arguments or results must be defined globally."
            )

    def __repr__(self) -> str:
        return f"<function {self.fullname}>"

    @property
    def fullname(self) -> str:
        """
        Return a fullname for a function including its module.
        """
        return f"{self.module}.{self.name}"

    def _calc_hash(self) -> str:
        # Hash based on fully qualified function name and source code.
        hash = hash_struct(
            [
                "Value.function",
                self.fullname.encode("utf8"),
                "source",
                get_func_source(self.instance),
            ]
        )
        return hash

    def __getstate__(self) -> dict:
        return {
            "hash": self._hash,
            "module": self.module,
            "name": self.name,
        }

    def __setstate__(self, state: dict) -> None:
        self._hash = state["hash"]
        self.module = state["module"]
        self.name = state["name"]
        try:
            # Try to reassociate the function based on its name.
            module = importlib.import_module(self.module)
            self.instance = getattr(module, self.name)
        except (ModuleNotFoundError, AttributeError):
            self.instance = make_unknown_function(self.fullname)

    def get_hash(self, data: Optional[bytes] = None) -> str:
        return self._hash

    def is_valid(self) -> bool:
        """
        Value is valid to use from the cache if we can reassociate the function.
        """
        return not is_unknown_function(self.instance)

    def serialize(self) -> bytes:
        # Serialize the Function wrapper.
        return pickle_dumps(self)

    @classmethod
    def deserialize(self, raw_type: Any, data: bytes) -> Any:
        # Deserialize the Function wrapper and return the inner python function.
        func_wrapper = pickle_loads(data)
        return func_wrapper.instance

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> Any:
        """
        Parses function by fully qualified name from command-line.
        """
        try:
            module_name, func_name = arg.rsplit(".", 1)
        except ValueError:
            raise ValueError(f"Unexpected format for function name: {arg}")

        try:
            module = importlib.import_module(module_name)
            func = getattr(module, func_name)
        except (ModuleNotFoundError, AttributeError):
            raise ValueError(f"Function not found: {arg}")

        return func


class FileCache(ProxyValue):
    """
    Baseclass for caching values to Files.

    This baseclass can be used to control where the serialized data of
    potentially large values is stored.

    .. code-block:: python

        # Example of user-specific data type.
        class Data:
            def __init__(self, data):
                self.data = data

        # Subclassing from FileCache and setting `type = Data` will register
        # with redun that File-caching should be used for any Value of type
        # `Data`.
        class DataType(FileCache):
            type = Data

            # This path specifies where the files containing serialized
            # values should be stored. This variable accepts all paths
            # acceptable to File (e.g. local, s3, etc).
            base_path = "tmp"

        @task()
        def task1() -> Data:
            # This value will be serialized to a File instead of the default
            # redun database.
            return Data("my data")
    """

    base_path = "."

    def _serialize(self) -> bytes:
        # User defined serialization.
        return pickle_dumps(self.instance)

    @classmethod
    def _deserialize(cls, data: bytes) -> Any:
        # User defined deserialization.
        return pickle_loads(data)

    def serialize(self) -> bytes:
        from redun.file import File

        # Serialize data.
        bytes = self._serialize()

        # Write (possibly large) data to file.
        filename = os.path.join(self.base_path, hash_bytes(bytes))
        File(filename).write(bytes, mode="wb")

        # Only cache filename.
        return filename.encode("utf8")

    @classmethod
    def deserialize(cls, raw_type: type, filename: bytes) -> Any:
        from redun.file import File

        # Read data from file.
        file = File(filename.decode("utf8"))
        if not file.exists():
            # File has since been delete. Treat as cache miss.
            raise InvalidValueError()

        data = cast(bytes, file.read("rb"))
        return cls._deserialize(data)



from redun.hashing import hash_struct, hash_tag_bytes


def test_hash_struct():
    """
    Test that we can hash a struct.
    """
    assert hash_struct(["Task", "123"]) == "f02cbd309f1866b7d015617265c233160abdbd10"
    assert hash_struct({"key1": 10, "key2": "123"}) == "15c28723885dd16585e5555596f5e0ba67ef9ce1"


def test_hash_tag_bytes():
    """
    Test that we can hash a tagged byte sequence.
    """
    assert hash_tag_bytes("Value", b"123") == "25aa6da917ca4e568e859df039fb1ca184e65d57"



import abc
import enum
import typing
from datetime import datetime
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple

from redun.handle import Handle
from redun.hashing import hash_struct
from redun.utils import MultiMap
from redun.value import TypeRegistry, get_type_registry

if typing.TYPE_CHECKING:
    from redun.scheduler import Job
    from redun.scheduler import Task as BaseTask

NULL = object()


def calc_call_hash(
    task_hash: str, args_hash: str, result_hash: str, child_call_hashes: List[str]
) -> str:
    return hash_struct(["CallNode", task_hash, args_hash, result_hash, sorted(child_call_hashes)])


# Types.
TagMap = MultiMap[str, Any]
KeyValue = Tuple[str, Any]


class TagEntityType(enum.Enum):
    Execution = "Execution"
    Job = "Job"
    CallNode = "CallNode"
    Task = "Task"
    Value = "Value"
    Null = "Null"


class RedunBackend(abc.ABC):
    def __init__(self, type_registry: Optional[TypeRegistry] = None):
        self.type_registry: TypeRegistry = type_registry or get_type_registry()

    def load(self, migrate: Optional[bool] = None) -> None:
        pass

    def calc_current_nodes(self, task_hashes: Iterable[str]) -> None:
        pass

    def record_call_node(
        self,
        task_name: str,
        task_hash: str,
        args_hash: str,
        expr_args: Tuple[Tuple, dict],
        eval_args: Tuple[Tuple, dict],
        result_hash: str,
        child_call_hashes: List[str],
    ) -> str:
        pass

    def record_args(
        self, call_hash: str, expr_args: Tuple[Tuple, dict], eval_args: Tuple[Tuple, dict]
    ) -> None:
        pass

    def record_value(self, value: Any, data=None) -> str:
        pass

    def get_call_hash(self, task_hash: str, args_hash: str) -> Optional[str]:
        pass

    def get_eval_cache(self, eval_hash: str) -> Tuple[Any, bool]:
        pass

    def set_eval_cache(
        self, eval_hash: str, task_hash: str, args_hash: str, value: Any, value_hash: str = None
    ):
        pass

    def get_cache(self, call_hash: str) -> Any:
        pass

    def explain_cache_miss(self, task: "BaseTask", args_hash: str) -> Optional[Dict[str, Any]]:
        pass

    def get_value(self, value_hash: str) -> Any:
        pass

    def advance_handle(self, handles: List[Handle], child_handle: Handle):
        pass

    def rollback_handle(self, handle: Handle) -> None:
        pass

    def is_valid_handle(self, handle: Handle) -> bool:
        pass

    def record_execution(self, args: List[str]) -> str:
        pass

    def record_job_start(self, job: "Job", now: Optional[datetime] = None) -> Any:
        pass

    def record_job_end(
        self, job: "Job", now: Optional[datetime] = None, status: Optional[str] = None
    ) -> None:
        pass

    def get_job(self, job_id: str) -> Optional[dict]:
        """
        Returns details for a Job.
        """
        pass

    def record_tags(
        self,
        entity_type: TagEntityType,
        entity_id: str,
        tags: Iterable[Tuple[str, Any]],
        parents: Iterable = (),
        update: bool = False,
    ):
        pass

    def delete_tags(
        self, entity_id: str, tags: Iterable[KeyValue], keys: Iterable[str] = ()
    ) -> List[Tuple[str, str, str, Any]]:
        pass

    def update_tags(
        self,
        entity_type: TagEntityType,
        entity_id: str,
        old_keys: Iterable[str],
        new_tags: Iterable[KeyValue],
    ) -> List[Tuple[str, str, str, Any]]:
        pass

    def get_tags(self, entity_ids: List[str]) -> Dict[str, TagMap]:
        pass

    def iter_record_ids(self, root_ids: Iterable[str]) -> Iterator[str]:
        pass

    def get_records(self, ids: Iterable[str], sorted=True) -> Iterable[dict]:
        pass

    def put_records(self, records: Iterable[dict]) -> int:
        pass



from random import random
from time import sleep

from redun.config import Config
from redun.scheduler import Scheduler
from redun.task import task

API_LIMIT = 10
MEMORY_LIMIT = 3
CONFIG_DICT = {"limits": {"api": str(API_LIMIT), "memory": str(MEMORY_LIMIT)}}


def test_scheduler_limits_config() -> None:
    """
    Scheduled should correctly interpret limits from config and should convert limits to ints.
    """
    scheduler = Scheduler(config=Config(CONFIG_DICT))
    assert scheduler.limits == {key: int(value) for key, value in CONFIG_DICT["limits"].items()}


def test_scheduler_limits_utils() -> None:
    """
    Test the limits utility methods on scheduler for consuming, releasing, and checking limits.
    """
    scheduler = Scheduler(config=Config(CONFIG_DICT))

    job_limits = {"api": 1, "memory": MEMORY_LIMIT}

    # Check that a job that should have enough resources to run is correctly deemed within limits.
    assert scheduler._is_job_within_limits(job_limits)

    # Consume all resources for one limit and make sure scheduler's internal accounting is correct.
    scheduler._consume_resources(job_limits)
    assert scheduler.limits_used == {"api": job_limits["api"], "memory": MEMORY_LIMIT}

    # The scheduler should now recognize that it cannot fit another job given consumed resources.
    assert not scheduler._is_job_within_limits(job_limits)

    # Now, let's release the limits from the job and confirm scheduler's accounting.
    scheduler._release_resources(job_limits)
    assert scheduler.limits_used == {limit_name: 0 for limit_name in CONFIG_DICT["limits"]}


class API:
    """
    Trivial mock of a throttled API.

    This class will raise an exception if connect() would cause current connection count to be
    higher than the maximum connection count allowed.
    """

    def __init__(self, max_connections):
        self.current_connections = 0
        self.max_connections = max_connections
        self.max_connections_exceeded = False

    def connect(self):
        if self.current_connections == self.max_connections:
            self.max_connections_exceeded = True

        self.current_connections += 1

    def close(self):
        self.current_connections -= 1
        # Just a sanity check to make sure we are not somehow getting to negative connection count.
        assert self.current_connections >= 0


def test_scheduler_respects_limits() -> None:
    """
    Used to check the the scheduler is deferring jobs and scheduling them as resources are freed.
    """

    @task(limits=["api"], cache=False)
    def limit_task(api, i):
        """
        Task to connect to the "API", sleep a bit, then disconnect from the "API".

        We include the `i` argument so that each task is different in the eyes of the scheduler. If
        we just passed api, the scheduler would see N of this task as only needing one execution.
        """
        api.connect()
        sleep(random())
        api.close()

    @task()
    def workflow(api):
        return [limit_task(api, i) for i in range(5)]

    # If we run a workflow with more tasks than the API allows, we should exceed the maximum
    # connections because the scheduler has the resource limit set so high..
    max_connections = 2
    api = API(max_connections)
    scheduler = Scheduler(config=Config({"limits": {"api": max_connections * 2}}))
    scheduler.run(workflow(api))
    assert api.max_connections_exceeded

    # Now, we can do the same experiment as above but we can use a scheduler without resource
    # limits defined. In this case, the limit should default to 1 and we should not exceed the
    # maximum connections of the API instance.
    api = API(max_connections)
    scheduler = Scheduler()
    scheduler.run(workflow(api))
    assert not api.max_connections_exceeded



import json
import os
import tempfile
import threading
import time
from collections import OrderedDict, deque
from typing import Any, Deque, Dict, Iterator, List, Optional, Tuple, Union, cast

from redun.executors import aws_utils
from redun.executors.base import Executor, register_executor
from redun.executors.code_packaging import create_zip, package_code, parse_code_package_config
from redun.executors.command import REDUN_PROG, REDUN_REQUIRED_VERSION
from redun.executors.scratch import (
    SCRATCH_ERROR,
    SCRATCH_INPUT,
    SCRATCH_OUTPUT,
    ExceptionNotFoundError,
    get_job_scratch_dir,
    get_job_scratch_file,
)
from redun.executors.scratch import parse_job_error as _parse_job_error
from redun.executors.scratch import parse_job_result
from redun.file import File
from redun.hashing import hash_stream, hash_text
from redun.scheduler import Job, Scheduler, Traceback
from redun.task import Task
from redun.utils import pickle_dump

ARGS = ["code", "script", "task", "input", "output", "error"]
VALID_GLUE_WORKERS = ["Standard", "G.1X", "G.2X"]
ONESHOT_FILE = "glue_oneshot.py.txt"

# AWS Glue job statuses
GLUE_JOB_STATUSES = aws_utils.JobStatus(
    all=["STARTING", "RUNNING", "STOPPING", "SUCCEEDED", "FAILED", "ERROR", "STOPPED", "TIMEOUT"],
    inflight=["STARTING", "RUNNING", "STOPPING"],
    pending=["STARTING"],
    success=["SUCCEEDED"],
    failure=["FAILED", "ERROR"],
    stopped=["STOPPED"],
    timeout=["TIMEOUT"],
)

# These packages are needed for the redun lib to work on glue.
# Versions should remain up to date with setup.py
DEFAULT_ADDITIONAL_PYTHON_MODULES = [
    "alembic>=1.4",
    "mako",
    "promise",
    "s3fs>=2021.11.1",
    "sqlalchemy>=1.3.17or<2",  # Don't use commas as AWS won't parse them.
]


def get_spark_history_dir(s3_scratch_prefix: str) -> str:
    """
    Returns s3 scratch path for Spark UI monitoring files.
    """
    return os.path.join(s3_scratch_prefix, "glue", "spark_history")


def get_glue_oneshot_scratch_file(s3_scratch_prefix: str, code_hash: str) -> str:
    """
    Returns s3 scratch path for a code package tar file.
    """
    return os.path.join(s3_scratch_prefix, "glue", f"oneshot-{code_hash}.py")


def get_redun_lib_scratch_file(s3_scratch_prefix: str, lib_hash: str) -> str:
    """
    Returns s3 scratch path for a code package tar file.
    """
    return os.path.join(s3_scratch_prefix, "glue", f"redun-{lib_hash}.zip")


def get_redun_lib_files() -> Iterator[str]:
    """
    Iterates through the files of the redun library.
    """
    exclude_dirs = ["__pycache__", "/tests/"]

    redun_module = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    for root, dirs, files in os.walk(redun_module):
        for file in files:
            full_filepath = os.path.join(root, file)
            if all(pattern not in full_filepath for pattern in exclude_dirs):
                yield full_filepath


def package_redun_lib(s3_scratch_prefix: str) -> File:
    """
    Package redun lib to S3.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        base_path = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        temp_path = os.path.join(tmpdir, "redun.zip")
        lib_files = get_redun_lib_files()
        temp_file = create_zip(temp_path, base_path, lib_files)

        with temp_file.open("rb") as infile:
            lib_hash = hash_stream(infile)
        lib_file = File(get_redun_lib_scratch_file(s3_scratch_prefix, lib_hash))
        if not lib_file.exists():
            temp_file.copy_to(lib_file)
    return lib_file


def get_or_create_glue_job_definition(
    script_location: str,
    role: str,
    temp_dir: str,
    extra_py_files: str,
    spark_history_dir: str,
    additional_python_modules: List[str] = DEFAULT_ADDITIONAL_PYTHON_MODULES,
    aws_region: str = aws_utils.DEFAULT_AWS_REGION,
) -> str:
    """
    Gets or creates an AWS Glue Job.

    Parameters
    ----------
    script_location : str
        S3 path of script that runs Spark job.

    role : str
        ARN of IAM role to associate with Glue job.

    temp_dir : str
        S3 path of scratch directory associated with job data and code.

    extra_py_files : str
        Comma separated S3 file paths for additional Python code needed for job.

    spark_history_dir : str
        S3 path where Spark event logs are stored.

    additional_python_modules : List[str]
        Python modules to be installed with pip before job start.

    aws_region : str
        AWS region in which to run the Glue job.

    Returns
    -------
    str
        Unique Glue job definition name.
    """
    client = aws_utils.get_aws_client("glue", aws_region=aws_region)

    # Define job definition.
    glue_job_def = dict(
        Description=f"Redun oneshot glue job {REDUN_REQUIRED_VERSION}",
        Role=role,
        ExecutionProperty={
            "MaxConcurrentRuns": 1000,
        },  # account-max value.
        Command={
            "Name": "glueetl",
            "ScriptLocation": script_location,
            "PythonVersion": "3",
        },
        DefaultArguments={
            "--TempDir": temp_dir,
            "--enable-s3-parquet-optimized-committer": "true",
            "--extra-py-files": extra_py_files,
            "--additional-python-modules": ",".join(additional_python_modules),
            "--job-bookmark-option": "job-bookmark-disable",
            "--job-language": "python",
            "--enable-spark-ui": "true",
            "--enable-job-insights": "true",
            "--spark-event-logs-path": spark_history_dir,
        },
        MaxRetries=0,
        NumberOfWorkers=2,  # Jobs will override this, so set to minimum value.
        WorkerType="Standard",
        GlueVersion="3.0",
        Timeout=2880,
    )
    glue_job_def_hash = hash_text(json.dumps(glue_job_def, sort_keys=True))
    glue_job_name = f"redun-glue-{glue_job_def_hash}"

    try:
        # See if job definition already exists.
        client.get_job(JobName=glue_job_name)
    except client.exceptions.EntityNotFoundException:
        # Create job definition.
        resp = client.create_job(Name=glue_job_name, **glue_job_def)
        assert resp["Name"] == glue_job_name

    return glue_job_name


def get_default_glue_service_role(
    account_num: Optional[str] = None, aws_region: str = aws_utils.DEFAULT_AWS_REGION
) -> str:
    """
    Returns the default Glue service role for the current account.
    """
    if not account_num:
        caller_id = aws_utils.get_aws_client("sts", aws_region=aws_region).get_caller_identity()
        account_num = caller_id["Account"]
    return f"arn:aws:iam::{account_num}:role/service-role/AWSGlueServiceRole"


@register_executor("aws_glue")
class AWSGlueExecutor(Executor):
    """
    A redun Executor for running jobs on AWS Glue.
    """

    def __init__(self, name: str, scheduler: Optional[Scheduler] = None, config=None):
        super().__init__(name, scheduler=scheduler)
        if config is None:
            raise ValueError("AWSGlueExecutor requires config.")

        # Required config
        self.s3_scratch_prefix = config["s3_scratch"]

        # Optional config
        self.aws_region = config.get("aws_region", aws_utils.get_default_region())
        self.role = config.get("role") or get_default_glue_service_role(aws_region=self.aws_region)
        self.code_package = parse_code_package_config(config)
        self.code_file: Optional[File] = None
        self.debug = config.getboolean("debug", fallback=False)
        self.interval = config.getfloat("job_monitor_interval", 10.0)
        self.retry_interval = config.getfloat("job_retry_interval", 60.0)
        self.glue_job_prefix = config.get("glue_job_prefix", REDUN_PROG.upper())
        self.glue_job_name: Optional[str] = None
        self.spark_history_dir = config.get(
            "spark_history_dir", get_spark_history_dir(self.s3_scratch_prefix)
        )

        # Default task options
        self.default_task_options = {
            "workers": config.getint("workers", 10),
            "worker_type": config.get("worker_type", "G.1X"),
            "timeout": config.getint("timeout", 2880),
            "role": config.get("role"),
            "additional_libs": [],
            "extra_files": [],
        }

        # Execution state.
        self.is_running = False
        self._monitor_thread = threading.Thread(target=self._monitor, daemon=False)
        self._submit_thread = threading.Thread(target=self._submission_thread, daemon=False)
        self.pending_glue_jobs: Deque[Tuple["Job", Tuple, Dict]] = deque()
        self.running_glue_jobs: Dict[str, "Job"] = OrderedDict()
        self.preexisting_glue_jobs: Dict[str, str] = {}  # Job hash -> Job ID
        self._oneshot_path: Optional[str] = None
        self.redun_zip_location: Optional[str] = None

    def gather_inflight_jobs(self) -> None:
        for run in self.get_jobs(statuses=GLUE_JOB_STATUSES.inflight):
            hash = run["Arguments"].get("--job-hash")
            self.preexisting_glue_jobs[hash] = run["Id"]

    def get_jobs(self, statuses: Optional[List[str]] = None) -> Iterator[dict]:
        """
        Gets all job runs with given status.
        """
        client = aws_utils.get_aws_client("glue", aws_region=self.aws_region)
        paginator = client.get_paginator("get_job_runs")

        for page in paginator.paginate(JobName=self.glue_job_name):
            for run in page["JobRuns"]:
                if statuses:
                    if run["JobRunState"] in statuses:
                        yield run
                else:
                    yield run

    def get_or_create_job_definition(self) -> None:
        """
        Get or create the default Glue job.
        """
        if not self._oneshot_path:
            # Copy Glue oneshot file to S3 with unique hash.
            oneshot_file = File(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), ONESHOT_FILE)
            )
            oneshot_hash = hash_text(cast(str, oneshot_file.read()))
            oneshot_s3_path = get_glue_oneshot_scratch_file(self.s3_scratch_prefix, oneshot_hash)
            oneshot_s3_file = File(oneshot_s3_path)
            if not oneshot_s3_file.exists():
                oneshot_file.copy_to(oneshot_s3_file)
            self._oneshot_path = oneshot_s3_path

        if not self.redun_zip_location:
            self.redun_zip_location = package_redun_lib(self.s3_scratch_prefix).path

        self.glue_job_name = get_or_create_glue_job_definition(
            script_location=self._oneshot_path,
            role=self.role,
            spark_history_dir=self.spark_history_dir,
            temp_dir=self.s3_scratch_prefix,
            extra_py_files=self.redun_zip_location,
            aws_region=self.aws_region,
        )

    def _start(self) -> None:
        """
        Starts monitoring thread
        """
        if not self.is_running:
            self.is_running = True

        if not self._monitor_thread.is_alive():
            self._monitor_thread = threading.Thread(target=self._monitor, daemon=False)
            self._monitor_thread.start()

        if not self._submit_thread.is_alive():
            self._submit_thread = threading.Thread(target=self._submission_thread, daemon=False)
            self._submit_thread.start()

    def _monitor(self) -> None:
        """Thread for monitoring running AWS Glue jobs."""
        assert self._scheduler
        assert self.glue_job_name

        try:
            while self.is_running and (self.running_glue_jobs or self.pending_glue_jobs):
                # Process running glue jobs
                jobs = glue_describe_jobs(
                    list(self.running_glue_jobs.keys()),
                    glue_job_name=self.glue_job_name,
                    aws_region=self.aws_region,
                )

                for job in jobs:
                    self._process_job_status(job)

                time.sleep(self.interval)

        except Exception as error:
            self._scheduler.reject_job(None, error)

        self.stop()

    def _submission_thread(self) -> None:
        """
        Thread for submitting AWS Glue jobs.

        Jobs are submitted in approximately the order in which they are started.
        Job submission may fail due to too many other running jobs or insufficient DPUs
        available. If submission fails, that job moves to the back of the pending jobs
        queue in case there are other jobs that need fewer resources that could be
        successfully submitted in the meantime. Once submission fails for 5 jobs in
        a row, we wait `self.retry_interval` seconds before submitting another job.
        """
        assert self._scheduler
        try:
            while self.is_running and self.pending_glue_jobs:
                fail_counter = 0
                while fail_counter < 5 and self.pending_glue_jobs:
                    job, args, kwargs = self.pending_glue_jobs.popleft()
                    job_id = self.submit_pending_job(job)

                    if job_id is None:
                        fail_counter += 1
                        self.pending_glue_jobs.append((job, args, kwargs))
                    else:
                        self.running_glue_jobs[job_id] = job
                        fail_counter = 0

                time.sleep(self.retry_interval)

        except Exception as error:
            self._scheduler.reject_job(None, error)

        # Thread can exit when there are no more pending jobs. That's okay,
        # as new job submissions will restart it.

    def stop(self) -> None:
        self.is_running = False

    def _process_job_status(self, job: dict) -> None:
        assert self._scheduler

        error: Optional[Exception] = None
        error_traceback: Optional[Traceback] = None
        result: Optional[Any] = None
        redun_job = self.running_glue_jobs.get(job["Id"])
        assert redun_job

        if job["JobRunState"] in GLUE_JOB_STATUSES.success:
            result, exists = parse_job_result(self.s3_scratch_prefix, redun_job)
            if exists:
                error = None
            else:
                error = FileNotFoundError(
                    get_job_scratch_file(self.s3_scratch_prefix, redun_job, SCRATCH_OUTPUT)
                )
                error_traceback = None

        elif job["JobRunState"] in GLUE_JOB_STATUSES.stopped:
            error = AWSGlueJobStoppedError("Job stopped by user.")

        elif job["JobRunState"] in GLUE_JOB_STATUSES.failure:
            # Errors may not be unpicklable out of the Glue environment, so use CloudWatch
            # job insights to get a traceback.
            if self.debug:
                error, error_traceback = parse_job_error(self.s3_scratch_prefix, redun_job, job)
            else:
                error = AWSGlueError("Glue job failed: see log info.")
                error_traceback = Traceback.from_error(error)

                error_traceback.logs = get_job_insight_traceback(
                    job_id=job["Id"],
                    log_group_name=job["LogGroupName"],
                    aws_region=self.aws_region,
                    max_results=100,
                )

        elif job["JobRunState"] in GLUE_JOB_STATUSES.timeout:
            error = AWSGlueJobTimeoutError(job.get("ErrorMessage", ""))
            error_traceback = Traceback.from_error(error)

        # Job is still running
        else:
            return

        self.running_glue_jobs.pop(job["Id"])
        if error:
            self._scheduler.reject_job(redun_job, error, error_traceback=error_traceback)
        else:
            self._scheduler.done_job(redun_job, result)

    def _get_job_options(self, job: Job) -> dict:
        """
        Determines task options for a job.
        """
        assert job.task

        task_options = dict(self.default_task_options)
        task_options.update(job.get_options())
        return task_options

    def submit(self, job: Job, args: Tuple, kwargs: dict) -> None:
        """
        Submit job to executor.
        """
        assert self._scheduler
        assert job.task

        # Check glue job definition exists. Otherwise, create it.
        if not self.glue_job_name:
            self.get_or_create_job_definition()

        # Gather inflight jobs if this is the first submission, using `is_running` as a
        # way of determining if this is the first submission or not. If we are already running,
        # then we know we have already had jobs submitted and done the inflight check so no
        # reason to do that again here.
        if not self.debug and not self.is_running:
            # Precompute existing inflight jobs for job reuniting.
            self.gather_inflight_jobs()

        # Package code if not already done
        if self.code_package is not False and self.code_file is None:
            code_package = self.code_package or {}
            assert isinstance(code_package, dict)
            self.code_file = package_code(self.s3_scratch_prefix, code_package, use_zip=True)

        # Determine job options
        task_options = self._get_job_options(job)

        # Determine if we can reunite with a previous Glue output or job.
        glue_job_id: Optional[str] = None
        use_cache = task_options.get("cache", True)
        if use_cache and job.eval_hash in self.preexisting_glue_jobs:
            assert self.glue_job_name
            glue_job_id = self.preexisting_glue_jobs.pop(job.eval_hash)

            # Make sure glue API still has a status on this job
            existing_job = next(
                glue_describe_jobs(
                    [glue_job_id], glue_job_name=self.glue_job_name, aws_region=self.aws_region
                )
            )

            if existing_job:
                glue_job_id = existing_job["Id"]
                assert glue_job_id  # for mypy
                self.running_glue_jobs[glue_job_id] = job
                self.log(
                    "reunite redun job {redun_job} with Glue job {glue_job}:\n".format(
                        redun_job=job.id, glue_job=glue_job_id
                    )
                )
            else:
                glue_job_id = None

        if glue_job_id is None:
            # Set up files and data for run.
            input_path = get_job_scratch_file(self.s3_scratch_prefix, job, SCRATCH_INPUT)
            input_file = File(input_path)
            with input_file.open("wb") as out:
                pickle_dump([args, kwargs], out)

            self.pending_glue_jobs.append((job, args, kwargs))

        self._start()

    def submit_pending_job(self, job: Job) -> Union[str, None]:
        """
        Returns true if job submission was successful
        """
        assert job.task
        assert self.glue_job_name
        assert self.redun_zip_location
        assert self.code_file
        client = aws_utils.get_aws_client("glue", aws_region=self.aws_region)

        try:
            glue_resp = submit_glue_job(
                job,
                job.task,
                glue_job_name=self.glue_job_name,
                redun_zip_location=self.redun_zip_location,
                s3_scratch_prefix=self.s3_scratch_prefix,
                job_options=self._get_job_options(job),
                code_file=self.code_file,
                aws_region=self.aws_region,
            )
        except client.exceptions.ConcurrentRunsExceededException:
            self.log("Too many concurrent runs of the glue job. Waiting for some to complete...")
            return None
        except client.exceptions.ResourceNumberLimitExceededException:
            self.log("No AWS DPUs available. Waiting for some to free up...")
            return None

        self.log(
            "submit redun job {redun_job} as {job_type} {glue_job}:\n"
            "  job_id          = {glue_job}\n"
            "  job_name        = {job_name}\n"
            "  s3_scratch_path = {job_dir}\n"
            "  retries         = {retries}\n".format(
                redun_job=job.id,
                glue_job=glue_resp["JobRunId"],
                job_name=self.glue_job_name,
                job_type="AWS Glue job",
                job_dir=get_job_scratch_dir(self.s3_scratch_prefix, job),
                retries=glue_resp.get("ResponseMetadata", {}).get("RetryAttempts"),
            )
        )
        return glue_resp["JobRunId"]


def submit_glue_job(
    job: Job,
    a_task: Task,
    s3_scratch_prefix: str,
    glue_job_name: str,
    redun_zip_location: str,
    code_file: File,
    job_options: dict = {},
    aws_region: str = aws_utils.DEFAULT_AWS_REGION,
) -> Dict[str, Any]:
    """
    Submits a redun task to AWS glue.

    Parameters
    ----------
    job : Job
        The :class:`redun.scheduler.Job` that is in the process of resolving the Task `a_task`.

    a_task : Task
        The :class:`redun.task.Task` to be run by the `job`.

    s3_scratch_prefix : str
        Prefix for S3 url where job files, including additional Python and data files, will reside.

    glue_job_name : str
        Name of Glue job definition to run.

    redun_zip_location : str
        S3 path of zipped redun source code.

    code_file : File
        Code to be executed as a redun workflow. Path within `redun.file.File` should be an S3 url.

    job_options : Dict[str, Any]
        Option map to configure job initialization and execution.

        Required fields
            `worker_type`: str
            `timeout`: int
            `workers`: int

        Optional fields
            `additional_libs`: List[str]
            `extra_py_files`: List[str]
            `extra_files`: List[str]

    aws_region : str
        AWS region in which to run the Glue job.

    Returns
    -------
    Dict[str, Any]
        Response contains `JobRunId` field.

    Raises
    ------
    ValueError
        If `worker_type` is not a valid predefined Glue worker type string.
    """
    input_path = get_job_scratch_file(s3_scratch_prefix, job, SCRATCH_INPUT)
    output_path = get_job_scratch_file(s3_scratch_prefix, job, SCRATCH_OUTPUT)
    error_path = get_job_scratch_file(s3_scratch_prefix, job, SCRATCH_ERROR)

    # Assemble job arguments
    assert job.eval_hash
    glue_args = {
        "--check-version": REDUN_REQUIRED_VERSION,
        "--input": input_path,
        "--output": output_path,
        "--script": a_task.load_module,
        "--task": a_task.fullname,
        "--error": error_path,
        "--job-hash": job.eval_hash,
    }

    # Comma separated string of Python modules to be installed with pip before job start.
    if job_options.get("additional_libs"):
        glue_args["--additional-python-modules"] = ",".join(
            DEFAULT_ADDITIONAL_PYTHON_MODULES + job_options["additional_libs"]
        )

    # Extra python and data files are specified as comma separated strings.
    # Files are first copied to S3, as Glue requires them to be there.

    # Extra py files will be in an importable location at job start.
    # They can be either importable zip files, or .py source files.
    # Redun and the code to run are provided as importable zip files.
    scratch_dir = get_job_scratch_dir(s3_scratch_prefix, job)

    glue_args["--extra-py-files"] = ",".join([redun_zip_location, code_file.path]) + ","
    if job_options.get("extra_py_files"):
        addl_files = [aws_utils.copy_to_s3(f, scratch_dir) for f in job_options["extra_py_files"]]
        glue_args["--extra-py-files"] += ",".join(addl_files)

    # Extra files will be available in job's $PWD.
    if job_options.get("extra_files"):
        glue_args["--extra-files"] = ",".join(
            aws_utils.copy_to_s3(f, scratch_dir) for f in job_options["extra_files"]
        )

    # Validate job options
    if job_options["worker_type"] not in VALID_GLUE_WORKERS:
        raise ValueError(f"Invalid worker type {job_options['worker_type']}")

    # Submit glue job
    # Any submission exceptions need to be handled by calling function.
    glue_client = aws_utils.get_aws_client("glue", aws_region=aws_region)
    result = glue_client.start_job_run(
        JobName=glue_job_name,
        Arguments=glue_args,
        Timeout=job_options["timeout"],
        WorkerType=job_options["worker_type"],
        NumberOfWorkers=job_options["workers"],
    )

    return result


def glue_describe_jobs(
    job_ids: List[str], glue_job_name: str, aws_region: str = aws_utils.DEFAULT_AWS_REGION
) -> Iterator[Dict[str, Any]]:

    glue_client = aws_utils.get_aws_client("glue", aws_region=aws_region)

    for id in job_ids:
        response = glue_client.get_job_run(
            JobName=glue_job_name, RunId=id, PredecessorsIncluded=False
        )
        yield response.get("JobRun")


def parse_job_error(
    s3_scratch_prefix: str, job: Job, glue_job_metadata: Optional[dict] = None
) -> Tuple[Exception, "Traceback"]:
    """
    Parse job error from s3 scratch path.
    """
    assert job.task

    error, error_traceback = _parse_job_error(s3_scratch_prefix, job)

    # Handle AWS Batch-specific errors.
    if isinstance(error, ExceptionNotFoundError) and glue_job_metadata:
        message = glue_job_metadata.get(
            "ErrorMessage", "Exception and traceback could not be found for AWS Glue Job"
        )
        error = AWSGlueError(message)
        error_traceback = Traceback.from_error(error)

    return error, error_traceback


def get_job_insight_traceback(
    job_id: str,
    log_group_name: str,
    aws_region: str = aws_utils.DEFAULT_AWS_REGION,
    max_results=200,
) -> List[str]:
    """
    Gets the traceback from AWS' Glue Job insights.
    """
    log_lines = aws_utils.iter_log_stream(
        log_group_name=f"{log_group_name}/logs-v2",
        log_stream=f"{job_id}-job-insights-rca-driver",
        limit=max_results,
        reverse=False,
        required=False,
        aws_region=aws_region,
    )

    if not log_lines:
        return ["No job insights found. See AWS Glue GUI online."]

    traceback = []
    for line in log_lines:
        message = line.get("message")
        if message and "Failure Reason" in message:
            # Message looks like "ERROR GlueException... [Glue Exception Analysis] {DICT we want}".
            try:
                traceback.extend(
                    json.loads(message[message.index("{") :])["Failure Reason"].split("\n")
                )
            except Exception:
                continue

    return traceback


class AWSGlueError(Exception):
    pass


class AWSGlueJobTimeoutError(Exception):
    """
    Custom exception to raise when AWS Glue jobs are killed due to timeout.
    """

    pass


class AWSGlueJobStoppedError(Exception):
    pass



import pickle
from typing import List
from unittest import mock

import pytest

from redun import Handle, Scheduler, merge_handles, namespace, task
from redun.backends.db import Handle as HandleRow
from redun.backends.db import RedunBackendDb
from redun.handle import create_handle
from redun.utils import pickle_dumps
from redun.value import get_type_registry


class DBConn:
    """
    Mock database connection class.
    """

    def __init__(self, uri):
        self.uri = uri

    def execute(self, sql):
        return [1]


class DbHandle(Handle):
    def __init__(self, name: str, db_file: str):
        self.db_file = db_file
        self.conn = DBConn(db_file)

    def execute(self, sql: str):
        return self.conn.execute(sql)


def get_handle(scheduler: Scheduler, conn: Handle) -> HandleRow:
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    return scheduler.backend.session.query(HandleRow).filter_by(hash=conn.__handle__.hash).one()


def test_handle_class() -> None:
    # Subclasses of Handle automatically register into registry.
    registry = get_type_registry()
    registry.parse_type_name("redun.tests.test_handle.DbHandle") == DbHandle

    # Handle subclass should have __handle_class__ info.
    assert DbHandle.type_name == "redun.tests.test_handle.DbHandle"


def test_handle_init() -> None:
    # Handle instances should have __handle__ info.
    conn = DbHandle("mydb", "data.db")
    assert conn.__handle__.name == "mydb"
    assert conn.__handle__.args == ("data.db",)
    assert conn.__handle__.kwargs == {}

    # Handle instance should also have their own custom data.
    assert conn.db_file == "data.db"
    assert isinstance(conn.conn, DBConn)

    # Pretend to use connection.
    assert conn.execute("SELECT 1") == [1]

    # Handles should support serialization.
    state: dict = {
        "name": "mydb",
        "namespace": "",
        "args": ("data.db",),
        "kwargs": {},
        "class_name": "redun.tests.test_handle.DbHandle",
        "call_hash": "",
        "key": "",
        "hash": "1810d6454a2f9706c3adb6cc2ac0c5b317c0757d",
    }
    assert conn.__getstate__() == state

    # Recreate the handle from serialized data.
    conn2 = DbHandle.__new__(DbHandle, state["name"], *state["args"], **state["kwargs"])
    conn2.__init__(state["name"], *state["args"], **state["kwargs"])  # type: ignore
    assert conn2.db_file == "data.db"
    assert isinstance(conn2.conn, DBConn)

    # Recreate the handle from serialized data like pickle does.
    conn3 = DbHandle.__new__(DbHandle)
    conn3.__setstate__(state)
    assert conn3.db_file == "data.db"
    assert isinstance(conn3.conn, DBConn)


def test_create_handle() -> None:
    state = {
        "name": "mydb",
        "args": ("data.db",),
        "kwargs": {},
        "class_name": "redun.tests.test_handle.DbHandle",
    }
    conn = create_handle(state)
    assert conn.db_file == "data.db"
    assert isinstance(conn.conn, DBConn)


def test_handle_pickle() -> None:
    """
    Handles should pickle correctly and recall the __init__ method.
    """
    conn = DbHandle("mydb", "data.db")
    data = pickle_dumps(conn)

    with mock.patch.object(DbHandle, "__init__") as init:
        init.return_value = None
        pickle.loads(data)
        init.assert_called_with("mydb", "data.db")


def test_handle_no_namespace(scheduler: Scheduler) -> None:
    @task()
    def task1():
        conn = DbHandle("conn", "data.db")
        return (conn.__handle__.namespace, conn.__handle__.name, conn.__handle__.fullname)

    assert scheduler.run(task1()) == ("", "conn", "conn")


# TODO: bring back.
def _test_handle_task_namespace():
    @task(namespace="myspace")
    def task1():
        conn = DbHandle("conn", "data.db")
        return (conn.__handle__.namespace, conn.__handle__.name, conn.__handle__.fullname)

    scheduler = Scheduler()
    assert scheduler.run(task1()) == ("myspace", "conn", "myspace.conn")


# TODO: bring back.
def _test_handle_scope_namespace():
    namespace("myspace")

    @task()
    def task1():
        conn = DbHandle("conn", "data.db")
        return (conn.__handle__.namespace, conn.__handle__.name, conn.__handle__.fullname)

    namespace()

    scheduler = Scheduler()
    assert scheduler.run(task1()) == ("myspace", "conn", "myspace.conn")


def test_handle_apply() -> None:
    conn = DbHandle("mydb", "data.db")
    conn2 = conn.apply_call("call_hash123")
    assert conn2.__handle__.hash != conn.__handle__.hash

    assert conn2.db_file == conn.db_file
    assert conn2.conn == conn.conn


def test_handle_fork() -> None:
    conn = DbHandle("mydb", "data.db")
    conn2 = conn.fork("1")
    assert conn2.__handle__.hash != conn.__handle__.hash

    assert conn2.db_file == conn.db_file
    assert conn2.conn == conn.conn


def test_handle_edges(scheduler: Scheduler) -> None:
    @task()
    def do_write(conn, data):
        conn.execute(data)
        return conn

    @task()
    def pipeline(conn):
        conn = do_write(conn, "a")
        conn = do_write(conn, "b")
        conn = do_write(conn, "c")
        return conn

    @task()
    def main():
        conn = DbHandle("conn", "postgresql://host/db")
        return pipeline(conn)

    scheduler.run(main())

    # We should be able to see both the pre and post processing handles.
    # 1 original + 3 do_writes * 2 + 1 pipeline
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    handles = scheduler.backend.session.query(HandleRow).all()
    assert len(handles) == 8


def test_handle_tasks_simple(scheduler: Scheduler) -> None:
    """
    Ensure a handle changes hash as it passes through tasks.
    """

    @task()
    def task1(conn):
        return conn

    # Pass handle through one task.
    conn1 = DbHandle("conn", "data.db")
    conn2 = scheduler.run(task1(conn1))

    # Handle hash should update.
    assert conn2.__handle__.hash != conn1.__handle__.hash

    # 3 handles should exist in HandleGraph (conn1, pre_conn1, conn2).
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    handles = scheduler.backend.session.query(HandleRow).all()
    assert len(handles) == 3

    # The should all be valid.
    assert all(handle.is_valid for handle in handles)

    # Let's update the task.
    @task()
    def task2(conn):
        return conn

    # Running the handle through the updated task should cause a branch in the HandleGraph.
    conn3 = scheduler.run(task2(conn1))
    handle3 = get_handle(scheduler, conn3)
    assert handle3.is_valid

    # conn2 should have rolled back.
    handle2 = get_handle(scheduler, conn2)
    assert not handle2.is_valid

    # conn3 should be a descendant of conn1.
    assert handle3.parents[0].parents[0].hash == conn1.__handle__.hash

    # Let's inspect the branch created in the HandleGraph.
    child_hashes = {h.hash: h for h in handle3.parents[0].children}
    assert conn3.__handle__.hash in child_hashes
    assert conn2.__handle__.hash in child_hashes

    # Rerunning task1 should revert the handle to its old state.
    conn2b = scheduler.run(task1(conn1))
    assert conn2b.__handle__.hash == conn2.__handle__.hash
    handle2 = get_handle(scheduler, conn2)
    handle2b = get_handle(scheduler, conn2b)
    assert handle2b.is_valid

    # conn3 should have rolled back.
    handle3 = get_handle(scheduler, conn3)
    assert not handle3.is_valid


def test_handle_tasks(scheduler: Scheduler) -> None:
    """
    Ensure a handle changes hash as it passes through tasks.
    """
    task_calls = []
    handles = {}

    @task()
    def init_db(conn: DbHandle):
        task_calls.append("init_db")
        handles["init_db"] = conn
        conn.execute("CREATE TABLE ...")
        return conn

    @task(version="v1")
    def update_db(conn: DbHandle):
        task_calls.append("update_db")
        handles["update_db"] = conn
        conn.execute("INSERT INTO ...")
        return conn

    @task()
    def workflow():
        task_calls.append("workflow")
        conn = DbHandle("conn", "data.db")
        handles["workflow"] = conn

        conn = init_db(conn)
        conn = update_db(conn)
        return conn

    final_conn = scheduler.run(workflow())

    # All the tasks should run in order.
    assert task_calls == ["workflow", "init_db", "update_db"]

    # We should get a changing handle hash.
    assert handles["workflow"].__handle__.hash != handles["init_db"].__handle__.hash
    assert handles["init_db"].__handle__.hash != handles["update_db"].__handle__.hash

    # Change the update_db task.
    @task(version="v2")  # type: ignore[no-redef]
    def update_db(conn: DbHandle):
        task_calls.append("update_db2")
        handles["update_db2"] = conn
        conn.execute("INSERT INTO ... something different;")
        return conn

    task_calls = []
    final_conn2 = scheduler.run(workflow())

    # Just the update_db task should run again.
    assert task_calls == ["update_db2"]

    # The previous final conn shouldn't be valid anymore, because
    # update_db2 branches off before it.
    final_handle = get_handle(scheduler, final_conn)
    final_handle2 = get_handle(scheduler, final_conn2)
    assert not final_handle.is_valid
    assert final_handle2.is_valid

    # The early handles used for init_db, should still be valid, since they
    # are ancestral to to final_handle2.
    assert scheduler.backend.is_valid_handle(handles["workflow"])
    assert scheduler.backend.is_valid_handle(handles["init_db"])
    assert scheduler.backend.is_valid_handle(handles["update_db"])
    assert scheduler.backend.is_valid_handle(handles["update_db2"])

    # The handle entering update_db should be the same hash.
    assert handles["update_db"].__handle__.hash == handles["update_db2"].__handle__.hash

    # Revert the update_db task.
    @task(version="v1")  # type: ignore[no-redef]
    def update_db(conn: DbHandle):
        task_calls.append("update_db")
        handles["update_db"] = conn
        conn.execute("INSERT INTO ...")
        return conn

    task_calls = []
    final_conn3 = scheduler.run(workflow())
    get_handle(scheduler, final_conn3)

    # Because we are using a handle, we should NOT have a fast revert (task_calls == []).
    # Instead update_db needs to run again.
    assert task_calls == ["update_db"]


def test_handle_rollback_forks(scheduler: Scheduler) -> None:
    """
    Ensure previous handles rollback.
    """
    task_calls = []
    handles = {}

    @task()
    def update_db(conn: DbHandle, data):
        task_calls.append(("update_db", data))
        handles["update_db", data] = conn
        conn.execute("INSERT INTO ...")
        return conn

    @task()
    def post_update_db(conn: DbHandle, data):
        handles["post_update_db", data] = conn
        return conn

    @task()
    def workflow():
        conn = DbHandle("conn", "data.db")
        handles["workflow"] = conn
        conn2 = post_update_db(update_db(conn, 10), 10)
        conn3 = post_update_db(update_db(conn, 20), 20)
        return merge_handles([conn2, conn3])

    scheduler.run(workflow())

    # Change the workflow.
    @task()  # type: ignore
    def workflow():
        conn = DbHandle("conn", "data.db")
        handles["workflow"] = conn
        conn2 = post_update_db(update_db(conn, 12), 12)
        conn3 = post_update_db(update_db(conn, 20), 20)
        return merge_handles([conn2, conn3])

    prev_handles = dict(handles)
    task_calls = []
    scheduler.run(workflow())

    # The first call to update_db(conn, 10) should have caused a rollback.
    assert not scheduler.backend.is_valid_handle(prev_handles["post_update_db", 10])
    assert scheduler.backend.is_valid_handle(handles["post_update_db", 12])

    # The second call, update_db(conn, 20), should be the same handle and valid.
    assert (
        prev_handles["post_update_db", 20].__handle__.hash
        == handles["post_update_db", 20].__handle__.hash
    )
    assert scheduler.backend.is_valid_handle(prev_handles["post_update_db", 20])


def test_handle_nested(scheduler: Scheduler) -> None:
    @task()
    def task1(conn1, conn2):
        return [conn1, conn2]

    conn1 = DbHandle("conn1", "data1.db")
    conn2 = DbHandle("conn2", "data2.db")

    conn1b, conn2b = scheduler.run(task1(conn1, conn2))

    # Handles in a nested structure should also advance their hash after task call.
    assert conn1.__handle__.hash != conn1b.__handle__.hash
    assert conn2.__handle__.hash != conn2b.__handle__.hash


def test_handle_merge(scheduler: Scheduler) -> None:
    """
    Test that we can merge two handles into one.
    """

    @task()
    def task1(conn):
        return conn

    @task()
    def task2(conn):
        return conn

    @task()
    def workflow():
        conn = DbHandle("conn", "data.db")
        conns = [
            task1(conn),
            task2(conn),
        ]
        return merge_handles(conns)

    conn = scheduler.run(workflow())

    # Assert that we have a handle with two parents in the handle graph.
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    handle = (
        scheduler.backend.session.query(HandleRow)
        .filter(HandleRow.hash == conn.__handle__.hash)
        .first()
    )
    assert len(handle.parents) == 2


def test_handle_merge3(scheduler: Scheduler) -> None:
    """
    Test that we can merge three handles into one.
    """

    @task()
    def task1(conn, arg):
        return conn

    @task()
    def workflow():
        conn = DbHandle("conn", "data.db")
        conns = [
            task1(conn, "a"),
            task1(conn, "b"),
            task1(conn, "c"),
        ]
        return merge_handles(conns)

    conn = scheduler.run(workflow())

    # Assert that we have a handle with two parents in the handle graph.
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    handle = (
        scheduler.backend.session.query(HandleRow)
        .filter(HandleRow.hash == conn.__handle__.hash)
        .first()
    )
    assert len(handle.parents) == 3


def test_handle_fork_many(scheduler: Scheduler) -> None:
    """
    Test that we can use named forking to allow narrow re-execution.
    """

    @task()
    def task1(conn, arg):
        task_calls.append(arg)
        return conn

    @task()
    def workflow():
        conn = DbHandle("conn", "data.db")
        conns = [
            task1(conn.fork("a"), "a"),
            task1(conn.fork("b"), "b"),
            task1(conn.fork("c"), "c"),
        ]
        return merge_handles(conns)

    task_calls: List[str] = []
    scheduler.run(workflow())
    assert set(task_calls) == {"a", "b", "c"}

    # Redefine workflow to change only order of parallel tasks.
    # We should not need to re-execute any of them.
    @task()  # type: ignore
    def workflow():
        conn = DbHandle("conn", "data.db")
        conns = [
            task1(conn.fork("b"), "b"),
            task1(conn.fork("c"), "c"),
            task1(conn.fork("a"), "a"),
        ]
        return merge_handles(conns)

    task_calls = []
    scheduler.run(workflow())
    assert task_calls == []


def test_handle_implicit_fork(scheduler: Scheduler) -> None:
    """
    Test that our heuristic for reconciling handles with implicit forks works.
    """

    @task()
    def task1(conn, arg):
        task_calls.append(arg)
        return conn

    @task(version="1")
    def task2(conn, arg):
        task_calls.append(arg)
        return conn

    @task()
    def workflow():
        conn = DbHandle("conn", "data.db")
        conns = [
            task1(conn, "a"),
            task2(conn, "b"),
            task1(conn, "c"),
        ]
        return merge_handles(conns)

    task_calls: List[str] = []
    scheduler.run(workflow())
    assert set(task_calls) == {"a", "b", "c"}

    # Executing again should be fully cached.
    task_calls = []
    scheduler.run(workflow())
    assert task_calls == []

    # Redfine workflow to just have more tasks at the end.
    # Just one more task should be executed.
    @task()  # type: ignore
    def workflow():
        conn = DbHandle("conn", "data.db")
        conns = [
            task1(conn, "a"),
            task2(conn, "b"),
            task1(conn, "c"),
            task1(conn, "d"),
        ]
        return merge_handles(conns)

    task_calls = []
    scheduler.run(workflow())
    assert set(task_calls) == {"d"}

    # Without explicit forks our heuristic (call order) for Handle
    # reconciliation isn't good enough. Let's just confirm that.
    @task(version="2")  # type: ignore[no-redef]
    def task2(conn, arg):
        task_calls.append(arg)
        return conn

    @task()  # type: ignore
    def workflow():
        conn = DbHandle("conn", "data.db")
        conns = [
            task2(conn, "b"),  # Updated task2.
            task1(conn, "c"),
            task1(conn, "a"),
        ]
        return merge_handles(conns)

    task_calls = []
    scheduler.run(workflow())
    assert set(task_calls) == {"a", "b", "c"}

    # Executing again should be fully cached.
    task_calls = []
    scheduler.run(workflow())
    assert task_calls == []


def test_handle_serial(scheduler: Scheduler) -> None:
    """
    Test that we can extend a serial pipeline with proper caching.
    """

    @task()
    def task1(conn, arg):
        task_calls.append(arg)
        return conn

    @task(version="1")
    def task2(conn, arg):
        task_calls.append(arg)
        return conn

    @task()
    def workflow():
        conn = DbHandle("conn", "data.db")
        conn = task1(conn, "a")
        conn = task2(conn, "b")
        conn = task1(conn, "c")
        return conn

    task_calls: List[str] = []
    scheduler.run(workflow())
    assert set(task_calls) == {"a", "b", "c"}

    # Executing again should be fully cached.
    task_calls = []
    scheduler.run(workflow())
    assert task_calls == []

    # Redfine workflow to just have more tasks at the end.
    # Just one more task should be executed.
    @task()  # type: ignore
    def workflow():
        conn = DbHandle("conn", "data.db")
        conn = task1(conn, "a")
        conn = task2(conn, "b")
        conn = task1(conn, "c")
        conn = task1(conn, "d")
        return conn

    task_calls = []
    scheduler.run(workflow())
    assert set(task_calls) == {"d"}


def test_handle_proxy() -> None:
    """
    Ensure that attributes access proxies to Handle.instance.
    """

    class DbHandle2(Handle):
        def __init__(self, name, db_file):
            self.instance = DBConn(db_file)

    conn = DbHandle2("mydb", "data.db")

    assert conn.uri == "data.db"
    assert conn.execute("select * from my_table;") == [1]
    assert conn.__handle__.name == "mydb"
    assert dict(conn.__getstate__(), hash="") == {
        "name": "mydb",
        "namespace": "",
        "args": ("data.db",),
        "kwargs": {},
        "class_name": "redun.tests.test_handle.DbHandle2",
        "call_hash": "",
        "key": "",
        "hash": "",
    }


def test_handle_proxy_attr_error() -> None:
    """
    Ensure that attributes don't proxy when Handle.instance doesn't exist.
    """

    class DbHandle2(Handle):
        def __init__(self, name, db_file):
            self.conn = DBConn(db_file)

    conn = DbHandle2("mydb", "data.db")

    assert hasattr(conn, "conn")
    assert not hasattr(conn, "uri")
    with pytest.raises(AttributeError):
        conn.uri


def test_multiple_forks(scheduler: Scheduler) -> None:
    """
    Ensure multiple handle forks are recorded.
    """

    @task()
    def task1(conn):
        return conn

    @task()
    def main():
        conn = DbHandle("conn", "data.db")
        conn = task1(conn)
        conn = task1(conn.fork("a").fork("b").fork("c"))
        return conn

    scheduler.run(main())

    # We should record both forks.
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    handles = scheduler.backend.session.query(HandleRow).all()
    assert len(handles) == 8



from typing import Dict, List, cast

from redun import File, Scheduler, task
from redun.backends.db import CallNode, Execution, RedunBackendDb, Value
from redun.backends.db.dataflow import (
    ArgumentValue,
    CallNodeValue,
    Task,
    display_dataflow,
    make_dataflow_dom,
    serialize_dataflow,
    walk_dataflow,
    walk_dataflow_argument_value,
    walk_dataflow_callnode,
    walk_dataflow_callnode_value,
    walk_dataflow_value,
)
from redun.tests.utils import Match, assert_match_text, use_tempdir


@use_tempdir
def test_dataflow_crawl(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Manually crawl through dataflow to ensure we can extract each kind of edge.
    """
    assert backend.session

    @task()
    def align_reads(fastq: File, ref_genome: File) -> File:
        reads = cast(str, fastq.read())
        ref = cast(str, ref_genome.read())
        bam = File(fastq.path.replace("fastq", "bam"))
        bam.write("align({}, {})".format(reads, ref))
        return bam

    @task()
    def call_variants(bam: File, ref_genome: File) -> File:
        align = cast(str, bam.read())
        ref = cast(str, ref_genome.read())
        vcf = File(bam.path.replace("bam", "vcf"))
        vcf.write("calls({}, {})".format(align, ref))
        return vcf

    @task()
    def align_reads_all(fastqs: List[File], ref_genome: File):
        bams = [align_reads(fastq, ref_genome) for fastq in fastqs]
        return bams

    @task()
    def call_variants_all(bams: List[File], ref_genome: File):
        vcfs = [call_variants(bam, ref_genome) for bam in bams]
        return vcfs

    @task()
    def main(fastqs: List[File], ref_genome: File):
        bams = align_reads_all(fastqs, ref_genome)
        vcfs = call_variants_all(bams, ref_genome)
        return vcfs

    # Create dataset.
    fastqs = []
    for i in range(5):
        fastq = File("sample{}.fastq".format(i))
        fastq.write("reads{}".format(i))
        fastqs.append(fastq)

    ref_genome = File("ref_genome")
    ref_genome.write("genome")

    # Run workflow.
    scheduler.run(main(fastqs, ref_genome))

    # Quick check workflow works as intended.
    assert File("sample4.vcf").read() == "calls(align(reads4, genome), genome)"

    vcf = File("sample4.vcf")
    value = backend.session.query(Value).filter(Value.value_hash == vcf.get_hash()).one()

    # Ensure we can walk to originating call nodes.
    edges = list(walk_dataflow_value(backend, value))
    assert [call_node.task_name for _, call_node in edges] == ["call_variants"]  # type: ignore

    # Ensure we can walk to arguments.
    [(_, call_node)] = edges
    assert isinstance(call_node, CallNode)
    edges2 = list(walk_dataflow_callnode(backend, call_node))

    # Ensure we can walk to upstream of arguments.
    [(_, bam_arg_val), (_, ref_arg_val)] = edges2
    assert isinstance(ref_arg_val, ArgumentValue)
    assert isinstance(bam_arg_val, ArgumentValue)
    edges3 = list(walk_dataflow_argument_value(backend, bam_arg_val))
    edges4 = list(walk_dataflow_argument_value(backend, ref_arg_val))

    # Ensure we can walk upstream of arguments again.
    [(_, bam_arg_val2)] = edges3
    [(_, ref_arg_val2)] = edges4
    assert isinstance(bam_arg_val2, ArgumentValue)
    edges5 = list(walk_dataflow_argument_value(backend, bam_arg_val2))
    # edges6 = list(walk_dataflow_argument_value(scheduler.backend, ref_arg_val2))

    # Walk bam back to specific align_reads() call node.
    [(_, bam_call_node_val)] = edges5
    assert isinstance(bam_call_node_val, CallNodeValue)
    edges7 = list(walk_dataflow_callnode_value(backend, bam_call_node_val))
    assert isinstance(edges7[0][1], CallNodeValue)
    assert edges7[0][1][1].file.path == "sample4.bam"

    # Walk bam back to fastq.
    [(_, bam_call_node_val2)] = edges7
    assert isinstance(bam_call_node_val2, CallNodeValue)
    edges8 = list(walk_dataflow_callnode_value(backend, bam_call_node_val2))

    [(_, align_node)] = edges8
    assert isinstance(align_node, CallNode)
    edges9 = list(walk_dataflow_callnode(backend, align_node))
    [fastq_edge, ref_edge] = edges9

    assert isinstance(fastq_edge[1], tuple)
    assert fastq_edge[1][1].file.path == "sample4.fastq"


@use_tempdir
def test_dataflow(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Automated dataflow extraction.
    """
    assert backend.session

    @task()
    def align_reads(fastq: File, ref_genome: File) -> File:
        reads = cast(str, fastq.read())
        ref = cast(str, ref_genome.read())
        bam = File(fastq.path.replace("fastq", "bam"))
        bam.write("align({}, {})".format(reads, ref))
        return bam

    @task()
    def call_variants(bam: File, ref_genome: File) -> File:
        align = cast(str, bam.read())
        ref = cast(str, ref_genome.read())
        vcf = File(bam.path.replace("bam", "vcf"))
        vcf.write("calls({}, {})".format(align, ref))
        return vcf

    @task()
    def align_reads_all(fastqs: List[File], ref_genome: File):
        bams = [align_reads(fastq, ref_genome) for fastq in fastqs]
        return bams

    @task()
    def call_variants_all(bams: List[File], ref_genome: File):
        vcfs = [call_variants(bam, ref_genome) for bam in bams]
        return vcfs

    @task()
    def main(fastqs: List[File], ref_genome: File):
        bams = align_reads_all(fastqs, ref_genome)
        vcfs = call_variants_all(bams, ref_genome)
        return vcfs

    # Create dataset.
    fastqs = []
    for i in range(5):
        fastq = File("sample{}.fastq".format(i))
        fastq.write("reads{}".format(i))
        fastqs.append(fastq)

    ref_genome = File("ref_genome")
    ref_genome.write("genome")

    # Run workflow.
    scheduler.run(main(fastqs, ref_genome))

    # Ensure workflow works as intended.
    assert File("sample4.vcf").read() == "calls(align(reads4, genome), genome)"

    vcf = File("sample4.vcf")
    value = backend.session.query(Value).filter(Value.value_hash == vcf.get_hash()).one()

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))

    # Render dataflow as text.
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <********> call_variants(bam, ref_genome)
  bam        = <********> File(path=sample4.bam, hash=********)
  ref_genome = <********> File(path=ref_genome, hash=********)

bam <-- argument of <********> call_variants_all(bams, ref_genome)
    <-- <********> align_reads_all(fastqs, ref_genome)
    <-- <********> align_reads(fastq, ref_genome_2)
  fastq        = <********> File(path=sample4.fastq, hash=********)
  ref_genome_2 = <********> File(path=ref_genome, hash=********)

fastq <-- argument of <********> align_reads_all(fastqs, ref_genome)
      <-- argument of <********> main(fastqs, ref_genome)
      <-- origin

ref_genome_2 <-- argument of <********> align_reads_all(fastqs, ref_genome)
             <-- argument of <********> main(fastqs, ref_genome)
             <-- <********> ref_genome_3

ref_genome <-- argument of <********> call_variants_all(bams, ref_genome)
           <-- argument of <********> main(fastqs, ref_genome)
           <-- <********> ref_genome_3

ref_genome_3 <-- origin
""",
        output,
    )

    # Serialize dataflow.
    dom_json = list(serialize_dataflow(dom))

    # Assert dataflow serialization.
    expected_json = [
        {
            "args": [
                {
                    "hash": Match(str),
                    "value_display": Match(regex=r"File\(path=sample4.bam, hash=........\)"),
                    "var_name": "bam",
                },
                {
                    "hash": Match(str),
                    "value_display": Match(regex=r"File\(path=ref_genome, hash=........\)"),
                    "var_name": "ref_genome",
                },
            ],
            "assign": {
                "display": "call_variants(bam, ref_genome)",
                "hash": Match(str),
                "job_id": Match(str),
                "execution_id": Match(str),
                "prefix": "",
                "var_name": "value",
            },
            "routing": [],
        },
        {
            "args": [
                {
                    "hash": Match(str),
                    "value_display": Match(regex=r"File\(path=sample4.fastq, hash=........\)"),
                    "var_name": "fastq",
                },
                {
                    "hash": Match(str),
                    "value_display": Match(regex=r"File\(path=ref_genome, hash=........\)"),
                    "var_name": "ref_genome_2",
                },
            ],
            "assign": {
                "display": "call_variants_all(bams, ref_genome)",
                "hash": Match(str),
                "job_id": Match(str),
                "execution_id": Match(str),
                "prefix": "argument of",
                "var_name": "bam",
            },
            "routing": [
                {
                    "display": "align_reads_all(fastqs, ref_genome)",
                    "hash": Match(str),
                    "job_id": Match(str),
                    "execution_id": Match(str),
                    "prefix": "",
                },
                {
                    "display": "align_reads(fastq, ref_genome_2)",
                    "hash": Match(str),
                    "job_id": Match(str),
                    "execution_id": Match(str),
                    "prefix": "",
                },
            ],
        },
        {
            "args": [],
            "assign": {
                "display": "align_reads_all(fastqs, ref_genome)",
                "execution_id": Match(str),
                "hash": Match(str),
                "job_id": Match(str),
                "prefix": "argument of",
                "var_name": "fastq",
            },
            "routing": [
                {
                    "display": "main(fastqs, ref_genome)",
                    "execution_id": Match(str),
                    "hash": Match(str),
                    "job_id": Match(str),
                    "prefix": "argument of",
                },
                {
                    "display": "origin",
                    "execution_id": None,
                    "hash": None,
                    "job_id": None,
                    "prefix": "",
                },
            ],
        },
        {
            "args": [],
            "assign": {
                "display": "align_reads_all(fastqs, ref_genome)",
                "execution_id": Match(str),
                "hash": Match(str),
                "job_id": Match(str),
                "prefix": "argument of",
                "var_name": "ref_genome_2",
            },
            "routing": [
                {
                    "display": "main(fastqs, ref_genome)",
                    "execution_id": Match(str),
                    "hash": Match(str),
                    "job_id": Match(str),
                    "prefix": "argument of",
                },
                {
                    "display": "ref_genome_3",
                    "execution_id": Match(str),
                    "hash": Match(str),
                    "job_id": Match(str),
                    "prefix": "",
                },
            ],
        },
        {
            "args": [],
            "assign": {
                "display": "call_variants_all(bams, ref_genome)",
                "execution_id": Match(str),
                "hash": Match(str),
                "job_id": Match(str),
                "prefix": "argument of",
                "var_name": "ref_genome",
            },
            "routing": [
                {
                    "display": "main(fastqs, ref_genome)",
                    "execution_id": Match(str),
                    "hash": Match(str),
                    "job_id": Match(str),
                    "prefix": "argument of",
                },
                {
                    "display": "ref_genome_3",
                    "execution_id": Match(str),
                    "hash": Match(str),
                    "job_id": Match(str),
                    "prefix": "",
                },
            ],
        },
        {
            "args": [],
            "assign": {
                "display": "origin",
                "execution_id": None,
                "hash": None,
                "job_id": None,
                "prefix": "",
                "var_name": "ref_genome_3",
            },
            "routing": [],
        },
    ]

    for i, (record, expected_record) in enumerate(zip(dom_json, expected_json)):
        assert record == expected_record


def test_dataflow_routing(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Dataflow display should indicate argument and result routing.
    """
    assert backend.session

    @task()
    def task2(x: int) -> int:
        return x + 1

    @task()
    def task1(x: int) -> int:
        return task2(x)

    @task()
    def task1b(x: int) -> int:
        return x + 1

    @task()
    def main(x: int) -> int:
        return task1b(task1(x))

    # Run workflow.
    assert scheduler.run(main(10)) == 12

    exec1 = backend.session.query(Execution).one()
    value = exec1.call_node.value

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <********> task1b(x)
  x = <********> 11

x <-- <********> task1(x)
  <-- <********> task2(x_2)
  x_2 = <********> 10

x_2 <-- argument of <********> task1(x)
    <-- argument of <********> main(x)
    <-- origin
""",
        output,
    )


@use_tempdir
def test_dataflow_subvalue(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Dataflow should handle tracing subvalues.
    """
    assert backend.session

    @task()
    def task1(x: int) -> File:
        file = File(str(x))
        file.write(str(x))
        return file

    @task()
    def task2(input: File) -> File:
        file = File(input.path + ".v2")
        file.write(input.read())
        return file

    @task()
    def process_files(files: List[File]) -> List[File]:
        return [task2(file) for file in files]

    @task()
    def main(x: int) -> List[File]:
        files = [task1(x), task1(x + 1)]
        return process_files(files)

    # Run workflow.
    assert len(scheduler.run(main(10))) == 2

    exec1 = backend.session.query(Execution).one()
    value = exec1.call_node.value

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <********> process_files(files)
  files = <********> [File(path=10, hash=********), File(path=11, hash=********)]

files <-- derives from
  task1_result   = <********> File(path=11, hash=********)
  task1_result_2 = <********> File(path=10, hash=********)

task1_result <-- <********> task1(x)
  x = <********> 11

task1_result_2 <-- <********> task1(x_2)
  x_2 = <********> 10

x_2 <-- argument of <********> main(x)
    <-- origin
""",
        output,
    )


@use_tempdir
def test_dataflow_subvalue_mix(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Dataflow should handle tracing subvalues that mix.
    """
    assert backend.session

    @task()
    def get_files() -> List[File]:
        file1 = File("a")
        file1.write("a")
        file2 = File("b")
        file2.write("b")
        return [file1, file2]

    @task()
    def process_files(files: List[File]) -> List[str]:
        return [cast(str, file.read()) for file in files]

    @task()
    def main() -> List[str]:
        files = get_files()
        return process_files([files[1], files[0]])

    # Run workflow.
    assert scheduler.run(main()) == ["b", "a"]

    exec1 = backend.session.query(Execution).one()
    value = exec1.call_node.value

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <********> process_files(files)
  files = <********> [File(path=b, hash=********), File(path=a, hash=********)]

files <-- derives from
  get_files_result = <********> [File(path=a, hash=********), File(path=b, hash=********)]

get_files_result <-- <********> get_files()
                 <-- origin
""",
        output,
    )


@use_tempdir
def test_dataflow_derives(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Dataflow should handle tracing subvalues that fanout.
    """
    assert backend.session

    @task()
    def list_files(dir):
        return ["prog.c", "lib.c"]

    @task()
    def compile(c_file):
        return c_file.replace(".c", ".o")

    @task()
    def link(o_files):
        return o_files[0][:-2]

    @task()
    def main(dir) -> str:
        c_files = list_files(dir)[:2]
        o_files = [compile(c_file) for c_file in c_files]
        prog = link(o_files)
        return prog

    # Run workflow.
    assert scheduler.run(main("my-prog/")) == "prog"

    exec1 = backend.session.query(Execution).one()
    value = exec1.call_node.value

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <48147ca5> link(o_files)
  o_files = <960234eb> ['prog.o', 'lib.o']

o_files <-- derives from
  compile_result   = <9cdc943d> 'lib.o'
  compile_result_2 = <5da457e9> 'prog.o'

compile_result <-- <a561a571> compile(c_file)
  c_file = <13c0193f> 'lib.c'

c_file <-- derives from
  list_files_result = <49a8fc1c> ['prog.c', 'lib.c']

compile_result_2 <-- <abf16029> compile(c_file_2)
  c_file_2 = <751a22a8> 'prog.c'

c_file_2 <-- derives from
  list_files_result = <49a8fc1c> ['prog.c', 'lib.c']

list_files_result <-- <b9dea9ed> list_files(dir)
  dir = <dc00c649> 'my-prog/'

dir <-- argument of <2151d173> main(dir)
    <-- origin
""",
        output,
    )


@use_tempdir
def test_dataflow_parallel(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Dataflow should handle tracing subvalues that fanout.
    """
    assert backend.session

    @task()
    def task1(x: int) -> int:
        return x + 1

    @task()
    def task2a(x: int) -> int:
        return x + 1

    @task()
    def task2b(x: int) -> int:
        return x + 1

    @task()
    def task3(a: int, b: int) -> int:
        return a + b

    @task()
    def main(x: int) -> int:
        y = task1(x)
        z1 = task2a(y)
        z2 = task2b(y)
        return task3(z1, z2)

    # Run workflow.
    assert scheduler.run(main(10)) == 24

    exec1 = backend.session.query(Execution).one()
    value = exec1.call_node.value

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <25136331> task3(a, b)
  a = <eb88d6a0> 12
  b = <eb88d6a0> 12

a <-- <d2cb0298> task2a(x)
  x = <bfe03b79> 11

b <-- <8410db36> task2b(x_2)
  x_2 = <bfe03b79> 11

x <-- <cf9b2b45> x_3

x_2 <-- <cf9b2b45> x_3

x_3 <-- <cf9b2b45> task1(x_4)
  x_4 = <279a8e0b> 10

x_4 <-- argument of <b08130cb> main(x)
    <-- origin
""",
        output,
    )


@use_tempdir
def test_dataflow_subvalue_parallel(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Dataflow should handle tracing subvalues that route in parallel.
    """
    assert backend.session

    File("file1").write("a")
    File("file2").write("b")

    @task()
    def list_files() -> Dict[str, File]:
        return {
            "file1": File("file1"),
            "file2": File("file2"),
        }

    @task()
    def process_files(file1: File, file2: File) -> str:
        return cast(str, file1.read()) + cast(str, file2.read())

    @task()
    def main() -> str:
        files = list_files()
        return process_files(files["file1"], files["file2"])

    # Run workflow.
    assert scheduler.run(main()) == "ab"

    exec1 = backend.session.query(Execution).one()
    value = exec1.call_node.value

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <********> process_files(file1, file2)
  file1 = <********> File(path=file1, hash=********)
  file2 = <********> File(path=file2, hash=********)

file1 <-- derives from
  list_files_result = <********> {'file1': File(path=file1, hash=********), 'file2': File(path=file2, hash=********)}

file2 <-- derives from
  list_files_result = <********> {'file1': File(path=file1, hash=********), 'file2': File(path=file2, hash=********)}

list_files_result <-- <********> list_files()
                  <-- origin
""",  # noqa: E501
        output,
    )


@use_tempdir
def test_default_argument_names(scheduler: Scheduler, backend: RedunBackendDb) -> None:
    """
    Dataflow should use the right argument names for parameters with default values.
    """

    @task()
    def get_planet() -> str:
        return "World"

    @task()
    def greeter(greet: str, thing: str) -> str:
        return "{}, {}!".format(greet, thing)

    @task()
    def main(greet: str = "Hello") -> str:
        return greeter(greet, get_planet())

    # Run workflow.
    assert scheduler.run(main()) == "Hello, World!"

    assert backend.session is not None
    exec1 = backend.session.query(Execution).one()
    value = exec1.call_node.value

    # Change the source code of a task.
    task1 = backend.session.query(Task).filter(Task.name == "greeter").one()
    task1.source = "incorrectly formatted function"
    backend.session.commit()

    # Extract dataflow.
    dataflow_edges = list(walk_dataflow(backend, value))
    dom = list(make_dataflow_dom(dataflow_edges))
    output = "\n".join(display_dataflow(dom))

    # Assert dataflow text.
    assert_match_text(
        """\
value <-- <fad2f196> greeter(arg0, arg1)
  arg0 = <d5d4dead> 'Hello'
  arg1 = <58dc3a20> 'World'

arg0 <-- argument of <2d6d02a4> main(greet)
     <-- origin

arg1 <-- <634ba277> get_planet()
     <-- origin
""",
        output,
    )



import importlib
import typing
from typing import Any, Callable, Dict, Iterator, Optional, Tuple, Type, cast

if typing.TYPE_CHECKING:
    from redun.scheduler import Job, Scheduler


class ExecutorError(Exception):
    pass


class Executor:
    def __init__(
        self,
        name: str,
        scheduler: Optional["Scheduler"] = None,
        config=None,
    ):
        self.name = name
        self._scheduler = scheduler

    def set_scheduler(self, scheduler: "Scheduler") -> None:
        self._scheduler = scheduler

    def log(self, *messages: Any, **kwargs) -> None:
        """
        Display log message through Scheduler.
        """
        assert self._scheduler
        self._scheduler.log(f"Executor[{self.name}]:", *messages, **kwargs)

    def submit(self, job: "Job", args: Tuple, kwargs: dict) -> None:
        assert self._scheduler
        self._scheduler.reject_job(
            job, ExecutorError("Executor {} does not support submitting tasks.".format(type(self)))
        )

    def submit_script(self, job: "Job", args: Tuple, kwargs: dict) -> None:
        assert self._scheduler
        self._scheduler.reject_job(
            job,
            ExecutorError(
                "Executor {} does not support submitting script tasks.".format(type(self))
            ),
        )

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass


# Singleton executor registry.
_executor_classes: Dict[str, Type[Executor]] = {}


def get_executor_class(executor_name: str, required: bool = True) -> Optional[Type[Executor]]:
    """
    Get an Executor by name from the executor registry.

    Parameters
    ----------
    executor_name : str
        Name of executor class to retrieve.
    required : bool
        If True, raises error if executor is not registered.
        If False, None is returned for unknown executor name.
    """
    executor_class = _executor_classes.get(executor_name)
    if required and not executor_class:
        raise ExecutorError("Unknown executor {}".format(executor_name))
    return executor_class


def _register_executor(executor_name: str, executor_class: Type[Executor]) -> None:
    """
    Register an Executor class to be used by the scheduler.
    """
    _executor_classes[executor_name] = executor_class


def register_executor(executor_name: str) -> Callable:
    """
    Register an Executor class to be used by the scheduler.
    """

    def deco(executor_class: Type[Executor]):
        _register_executor(executor_name, executor_class)
        return executor_class

    return deco


def get_executors_from_config(executors_config: dict) -> Iterator[Executor]:
    """
    Instantiate executors defined in an executors config section.
    """
    for executor_name, executor_config in executors_config.items():
        executor_class = cast(Type[Executor], get_executor_class(executor_config["type"]))
        executor = executor_class(executor_name, config=executor_config)
        yield executor


def load_task_module(module_name: str, task_name: str) -> None:
    """
    Helper method that Executors may call to load a task's module code.

    Presently, the main benefit is this method provides better error handling.

    Args:
          module_name: Task module name.
          task_name: Task name
    """
    try:
        importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        # Append user-friendly error to original exception
        raise ModuleNotFoundError(
            f"Failed to find module {module_name} for task {task_name}. "
            "If this is remotely executed, ensure module is included in code packaging."
        ) from exc



from typing import List, cast

import pytest

from redun import Scheduler, task
from redun.expression import Expression, TaskExpression
from redun.functools import compose
from redun.task import PartialTask
from redun.value import get_type_registry

NULL = object()


@task(name="map", namespace="redun")
def task_map(a_task, values):
    return [a_task(value) for value in values]


# Monkey patch Expression with map.
def expression_map(self, a_task):
    return TaskExpression("redun.map", (a_task, self), {})


Expression.map = expression_map  # type: ignore


@task(name="reduce", namespace="redun")
def task_reduce(a_task, values, init=NULL):
    values = iter(values)
    if init is NULL:
        result = next(values)
    else:
        result = init
    for value in values:
        result = a_task(result, value)
    return result


def expression_reduce(self, a_task, init=NULL):
    return TaskExpression("redun.reduce", (a_task, self, init), {})


Expression.reduce = expression_reduce  # type: ignore


def test_expression_map(scheduler):
    @task()
    def get_list():
        return [1, 2, 3]

    @task()
    def inc(x):
        return x + 1

    assert scheduler.run(get_list().map(inc)) == [2, 3, 4]


def test_expression_reduce(scheduler):
    @task()
    def get_list():
        return [1, 2, 3]

    @task()
    def add(a, b):
        return a + b

    assert scheduler.run(get_list().reduce(add)) == 6
    assert scheduler.run(get_list().reduce(add, 10)) == 16


def test_partial(scheduler: Scheduler) -> None:
    """
    PartialTask should be created correctly and callable.
    """

    @task()
    def add(a: int, b: int) -> int:
        return a + b

    assert scheduler.run(add(1, 2)) == 3

    # Create PartialTask and call it.
    partial_task: PartialTask = PartialTask(add, (1,), {})
    expr: Expression[int] = partial_task(2)
    assert isinstance(expr, TaskExpression)
    assert expr.args == (1, 2)
    assert scheduler.run(expr) == 3

    # Use Task.partial() to make PartialTask.
    inc = add.partial(1)
    assert scheduler.run(inc(10)) == 11

    # Partial should be usable in a workflow.
    @task()
    def main() -> List[int]:
        inc = add.partial(1)
        return [inc(10), inc(12)]

    assert scheduler.run(main()) == [11, 13]

    # Arguments can be specified in different order using kwargs.
    assert scheduler.run(add.partial(b=1)(10)) == 11
    assert scheduler.run(add.partial(a=1)(b=10)) == 11


def test_partial_serialize() -> None:
    """
    PartialTask should serialize and deserialize.
    """

    @task(name="my_add", namespace="my_namespace")
    def add(a: int, b: int) -> int:
        return a + b

    inc = add.partial(1)
    state = inc.__getstate__()
    assert state == {
        "args": (1,),
        "compat": [],
        "hash": "13b99deec09100b663f0ab0d0e9af49ee0a800e0",
        "kwargs": {},
        "name": "my_add",
        "namespace": "my_namespace",
        "script": False,
        "task": {
            "compat": [],
            "hash": "7bc36e593df9b494ba5cb47be1664c04d12ae851",
            "name": "my_add",
            "namespace": "my_namespace",
            "script": False,
            "task_options": {},
            "version": None,
        },
        "task_options": {},
        "version": None,
    }

    # Round trip deserialization.
    inc2 = PartialTask.__new__(PartialTask)
    inc2.__setstate__(state)
    assert inc2.__getstate__() == state
    assert inc2.source == add.source


def test_partial_bad_args(scheduler: Scheduler) -> None:
    """
    Scheduler should be able to catch and properly record bad arguments.
    """

    @task()
    def add3(a, b, c):
        return a + b + c

    @task()
    def main():
        # Specify argument 'a' twice. Leave argument 'c' unspecified.
        part = add3.partial(a=1, b=2)
        return part(3)

    with pytest.raises(TypeError):
        scheduler.run(main())


def test_partial_options(scheduler: Scheduler) -> None:
    """
    PartialTask should be able to use task options.
    """

    @task()
    def add(a: int, b: int) -> int:
        return a + b

    assert not add.has_task_option("memory")
    assert add.get_task_option("memory") is None
    assert 1 == add.get_task_option("memory", 1)

    inc = add.partial(1).options(memory=4)
    inc = cast(PartialTask, inc)

    # Should be recorded on task.
    assert inc.task.has_task_option("memory")
    assert 4 == inc.task.get_task_option("memory")
    assert 4 == inc.task.get_task_option("memory", 1)

    # When we call the PartialTask, the task_options should propagate to the Expression.
    expr: Expression[int] = inc(2)
    assert expr.task_expr_options == {"memory": 4}

    # Task options should serialize.
    state = inc.__getstate__()
    assert state["task"]["task_options"] == {"memory": 4}

    # Round trip deserialization.
    inc2 = PartialTask.__new__(PartialTask)
    inc2.__setstate__(state)
    assert inc2.__getstate__() == state


def test_partial_partial(scheduler: Scheduler) -> None:
    """
    We should be able to perform chained partial application.
    """

    @task()
    def add3(a, b, c):
        return a + b + c

    part = add3.partial(1).partial(2)
    assert part.args == (1, 2)
    assert scheduler.run(part(3)) == 6

    part2 = add3.partial(b=2).partial(a=1)
    assert part2.args == ()
    assert part2.kwargs == {"a": 1, "b": 2}
    assert scheduler.run(part2(c=3)) == 6

    # Serialize and deserialize PartialTask (simulate fetching from cache).
    registry = get_type_registry()
    data = registry.serialize(part)
    part3 = registry.deserialize("redun.PartialTask", data)

    # Should get same result when calling.
    assert scheduler.run(part3(3)) == 6


def test_partial_partial_options(scheduler: Scheduler) -> None:
    """
    We should be able to mix partial and options calls.
    """

    @task()
    def add3(a, b, c):
        return a + b + c

    # Options should be present on PartialTask.
    part = add3.partial(1).partial(2).options(memory=4)
    part = cast(PartialTask, part)
    assert part.task.get_task_options() == {"memory": 4}

    # Options should propagate to Expression.
    expr: Expression[int] = part(3)
    assert expr.task_expr_options == {"memory": 4}


def test_partial_as_arg():
    task_calls = []

    @task()
    def add(a, b):
        task_calls.append("add")
        return a + b

    @task()
    def main(a_task):
        task_calls.append("main")
        return a_task(10)

    scheduler = Scheduler()
    inc = add.partial(1)
    assert scheduler.run(main(inc)) == 11
    assert task_calls == ["main", "add"]

    # Everything should be cached.
    task_calls = []
    assert scheduler.run(main(inc)) == 11
    assert task_calls == []


def test_partial_as_result():
    task_calls = []

    @task()
    def add(a, b):
        task_calls.append("add")
        return a + b

    @task()
    def make_partial():
        task_calls.append("make_partial")
        inc = add.partial(1)
        return inc

    @task()
    def main():
        task_calls.append("main")
        inc = make_partial()
        return inc(10)

    scheduler = Scheduler()
    task_calls = []
    assert scheduler.run(main()) == 11
    assert task_calls == ["main", "make_partial", "add"]

    # Everything should be cached.
    task_calls = []
    assert scheduler.run(main()) == 11
    assert task_calls == []


def test_compose():
    @task()
    def task1(x):
        return x + 1

    @task()
    def task2(x):
        return x * 2

    scheduler = Scheduler()

    task3 = compose(task2, task1)
    assert scheduler.run(task3(1)) == 4

    task4 = compose(task1, task2, task1)
    assert scheduler.run(task4(1)) == 5


def test_lisp(scheduler):
    @task()
    def pair(a, b, attr):
        if attr == "head":
            return a
        elif attr == "tail":
            return b
        else:
            raise ValueError()

    @task()
    def add(a, b):
        return a + b

    @task()
    def identity(x):
        return x

    def defer(x):
        return identity.partial(x)

    def cons(a, b):
        return pair.partial(a, b)

    @task()
    def head(lst):
        return lst("head")

    @task()
    def tail(lst):
        return lst("tail")

    @task()
    def is_null(value):
        return value is None

    @task()
    def _cond_eval(expr, result1, result2):
        if expr:
            return result1()
        else:
            return result2()

    def cond(expr, result1, result2):
        return _cond_eval(expr, defer(result1), defer(result2))

    @task()
    def map_(func, lst):
        return cond(is_null(lst), None, cons(func(head(lst)), map_(func, tail(lst))))

    @task()
    def reduce(func, lst, init):
        return cond(is_null(lst), init, func(head(lst), reduce(func, tail(lst), init)))

    @task()
    def as_list(lst):
        return cond(is_null(lst), [], add([head(lst)], as_list(tail(lst))))

    @task()
    def _sequence(low, high, attr):
        if attr == "head":
            return low
        elif attr == "tail":
            if low + 1 < high:
                return sequence(low + 1, high)
            else:
                return None
        else:
            return ValueError()

    def sequence(low, high):
        return _sequence.partial(low, high)

    @task()
    def take(k, lst):
        return cond(is_null(lst).__or__(k <= 0), None, cons(head(lst), take(k - 1, tail(lst))))

    lst = cons(1, cons(2, cons(3, None)))

    assert scheduler.run(head(lst)) == 1
    assert scheduler.run(head(tail(lst))) == 2
    assert scheduler.run(cond(identity(True), identity(1), identity(2))) == 1
    assert scheduler.run(cond(identity(False), identity(1), identity(2))) == 2

    assert scheduler.run(as_list(map_(add.partial(1), lst))) == [2, 3, 4]
    assert scheduler.run(reduce(add, lst, 0)) == 6
    # assert scheduler.run(reduce(add, map_(list, lst), [])) == [1, 2, 3]

    assert scheduler.run(as_list(sequence(1, 10))) == [1, 2, 3, 4, 5, 6, 7, 8, 9]

    assert scheduler.run(as_list(take(5, sequence(1, 10)))) == [1, 2, 3, 4, 5]


def test_seq(scheduler):
    task_calls = []

    @task()
    def identity(x):
        return x

    def defer(x):
        return identity.partial(x)

    @task()
    def task1(x):
        task_calls.append(x)
        return x

    @task()
    def seq_(t1, *rest):
        if rest:
            return seq_(rest[0](), *rest[1:])
        return t1

    def seq(t1, *rest):
        return seq_(t1, *map(defer, rest))

    scheduler.run(seq(task1("a"), task1("b")))
    assert task_calls == ["a", "b"]



_current_namespace = ""


def namespace(_namespace=None):
    """
    Set the current task namespace.

    Use None to unset the current namespace.
    """
    global _current_namespace
    _current_namespace = _namespace or ""


def get_current_namespace():
    """
    Returns the current task namespace during import.
    """
    return _current_namespace



import inspect
import re
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from redun.expression import SchedulerExpression, TaskExpression
from redun.hashing import hash_arguments, hash_struct
from redun.namespace import get_current_namespace
from redun.promise import Promise
from redun.utils import get_func_source
from redun.value import Value, get_type_registry

Func = TypeVar("Func", bound=Callable)
Func2 = TypeVar("Func2", bound=Callable)
Result = TypeVar("Result")


def get_task_registry():
    """
    Returns the global task registry.
    """
    return _task_registry


def undefined_task(fullname: str, *args: Any, **kwargs: Any) -> None:
    """
    Default function used for a deserialized Task with unknown definition.
    """
    raise ValueError("Task {} is undefined.".format(fullname))


def get_tuple_type_length(tuple_type: Any) -> Optional[int]:
    """
    Returns the length of a tuple type if inferable.
    """
    if getattr(tuple_type, "__origin__", None) in (tuple, Tuple):
        # Return type is Tuple[ * ].

        # __args__ is not available on Tuple type in Python 3.9+ because it was removed from
        # _SpecialGenericAlias in:
        #
        #       https://github.com/python/cpython/pull/19984
        #
        # For more info, see bpo-40397 here:
        #
        #       https://bugs.python.org/issue40397
        #
        # typing.get_args was added in Python 3.8 so we can use that instead if we detect we are
        # running on Python 3.8+
        if sys.version_info >= (3, 8):
            from typing import get_args

            tuple_type_args = get_args(tuple_type)
        else:
            tuple_type_args = tuple_type.__args__

        if Ellipsis in tuple_type_args or len(tuple_type_args) == 0:
            # Tuple of unknown length.
            return None

        if tuple_type_args == ((),):
            # Special case for length zero.
            return 0

        return len(tuple_type_args)

    if inspect.isclass(tuple_type) and issubclass(tuple_type, tuple):
        # Return type is namedtuple.
        fields = getattr(tuple_type, "_fields", None)
        if fields:
            return len(fields)

    return None


class Task(Value, Generic[Func]):
    """
    A redun Task.

    Tasks are the basic unit of execution in redun. Tasks are often defined
    using the :func:`redun.task.task` decorator:

    .. code-block:: python

        @task()
        def my_task(x: int) -> int:
            return x + 1

    Similar to pickling of functions, Tasks specify the work to execute by reference, not by
    value. This is important for serialization and caching, since a task is always reattached
    to the latest implementation and task options just-in-time.

    Task options may give direction to the scheduler about the execution environment, but must
    not alter the result of the function, since changing options will not trigger cache
    invalidation and we are allowed to return cached results with differing options. These
    can be provided at two times: 1) at module-load time by decorator options, 2) at run-time
    by creating task expressions. The former are not serialized so they are attached the latest
    value in the source code. The latter must be serialized, since they are not available
    except at run-time.

    Task are routinely serialized and deserialized and are cached based on their serialization.
    This relies on the serialized format being as minimal as possible, while providing the needed
    reference to the code to run.

    If needed, extra hashable data may be provided to trigger re-evaluation/cache-invalidation.
    """

    type_name = "redun.Task"

    def __init__(
        self,
        func: Callable,
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[str] = None,
        compat: Optional[List[str]] = None,
        script: bool = False,
        task_options_base: Optional[dict] = None,
        task_options_override: Optional[dict] = None,
        hash_includes: Optional[list] = None,
        source: Optional[str] = None,
    ):
        self.name = name or func.__name__
        self.namespace = self._compute_namespace(func, namespace)
        self.func = func
        if source is not None:
            self.source = source
        else:
            self.source = get_func_source(func)

        self.version = version
        self.compat = compat or []
        self.script = script
        # The base set of options, typically defined by module-load time code, which are not
        # serialized and are not hashed
        self._task_options_base = task_options_base or {}
        # Overrides to the options, typically provided at run-time, which are serialized and
        # hashed.
        self._task_options_override = task_options_override or {}
        self._signature: Optional[inspect.Signature] = None
        # Extra data to hash, but not serialize
        self._hash_includes = hash_includes
        self.recompute_hash()

        self._validate()

    def recompute_hash(self):
        self.hash = self._calc_hash()

    @staticmethod
    def _compute_namespace(func: Callable, namespace: Optional[str] = None):
        """Compute the namespace for the provided function to wrap.

        WARNING: The computation is stateful and hence this can only be used during task
        creation.

        Precedence:
        - Explicit namespace provided (empty string is a valid explicit value)
        - Infer it from a `redun_namespace` variable in the same module as func
        - The current namespace, configured with `set_current_namespace`"""

        # Determine task namespace.
        if namespace is None:
            namespace = getattr(sys.modules[func.__module__], "redun_namespace", None)

        if namespace is not None:
            return namespace
        return get_current_namespace()

    @property
    def nout(self) -> Optional[int]:
        """
        Determines nout from task options and return type.

        The precedence is:
        - the task option
        - function return type
        """
        if self.has_task_option("nout"):
            nout = self.get_task_option("nout")
        else:
            # Infer nout from return type.
            return_type = self.func.__annotations__.get("return")
            nout = get_tuple_type_length(return_type)

        return nout

    T = TypeVar("T")

    @overload
    def get_task_option(self, option_name: str) -> Optional[Any]:
        ...

    @overload
    def get_task_option(self, option_name: str, default: T) -> T:
        ...

    def get_task_option(self, option_name: str, default: Optional[T] = None) -> Optional[T]:
        """
        Fetch the requested option, preferring run-time updates over options from task
        construction. Like the dictionary `get` method, returns the default
        a `KeyError` on missing keys.
        """

        if option_name in self._task_options_override:
            return self._task_options_override[option_name]
        if option_name in self._task_options_base:
            return self._task_options_base[option_name]
        return default

    def get_task_options(self) -> dict:
        """
        Merge and return the task options.
        """
        return {
            **self._task_options_base,
            **self._task_options_override,
        }

    def has_task_option(self, option_name: str) -> bool:
        """
        Return true if the task has an option with name `option_name`
        """
        return option_name in self._task_options_override or option_name in self._task_options_base

    def __repr__(self) -> str:
        return "Task(fullname={fullname}, hash={hash})".format(
            fullname=self.fullname,
            hash=self.hash[:8],
        )

    def _validate(self) -> None:
        config_args = self.get_task_option("config_args")
        if config_args:
            valid_args = list(self.signature.parameters)
            invalid_args = [
                config_arg for config_arg in config_args if config_arg not in valid_args
            ]
            if invalid_args:
                raise ValueError(
                    f"Invalid config args: {', '.join(invalid_args)}. Expected all config_args to "
                    f"be one of the function args/kwargs: {', '.join(valid_args)}"
                )

        if self.namespace:
            if not re.match("^[A-Za-z_][A-Za-z_0-9.]*$", self.namespace):
                raise ValueError(
                    f"Task namespace must use only alphanumeric characters, "
                    f"underscore '_', and dot '.', and may not start with a dot: {self.namespace}"
                )

        if not re.match("^[A-Za-z_][A-Za-z_0-9]*$", self.name):
            raise ValueError(
                f"Task name must use only alphanumeric characters and underscore '_': {self.name}"
            )

        # Validate nout.
        if self.nout is not None:
            if not isinstance(self.nout, int):
                raise TypeError("nout must be an int")
            if self.nout < 0:
                raise TypeError("nout must be non-negative")

    @property
    def fullname(self) -> str:
        """
        Returns the fullname of a Task: '{namespace}.{name}'
        """
        return self._format_fullname(self.namespace, self.name)

    @staticmethod
    def _format_fullname(namespace: Optional[str], name: str):
        """Format the fullname of a Task."""
        if namespace:
            return namespace + "." + name
        else:
            return name

    def _call(self: "Task[Callable[..., Result]]", *args: Any, **kwargs: Any) -> Result:
        """
        Returns a lazy Expression of calling the task.
        """
        return cast(
            Result,
            TaskExpression(
                self.fullname, args, kwargs, self._task_options_override, length=self.nout
            ),
        )

    # Typing strategy: Ideally, we could use more honest types for the arguments
    # and return type of `Task.__call__()`, such as `Union[Arg, Expression[Arg]]`
    # and `TaskExpression[Result]` respectively. However, dynamically creating
    # such a signature has several challenges at the moment, such as requiring a
    # mypy plugin and forcing users to wrap every task return type. Therefore, we
    # compromise and force cast `Task.__call__()` to have the same signature as
    # the wrapped function, `Func`. This allows users to write tasks with very
    # natural types, and mypy can catch most type errors. The one trade off is
    # that this approach is too permissive about using `TaskExpression[T]`
    # wherever `T` is allowed.

    # Cast the signature to match the wrapped function.
    __call__: Func = cast(Func, _call)

    def options(self, **task_options_update: Any) -> "Task[Func]":
        """
        Returns a new Task with task_option overrides.
        """
        new_task_options_update = {
            **self._task_options_override,
            **task_options_update,
        }
        return Task(
            self.func,
            name=self.name,
            namespace=self.namespace,
            version=self.version,
            compat=self.compat,
            script=self.script,
            source=self.source,
            task_options_base=self._task_options_base,
            task_options_override=new_task_options_update,
        )

    def _calc_hash(self) -> str:
        """The hash is designed for checking equality of `Task`s for the purposes of caching.
        That is, results of this Task may be cached if the hash value is the same, regardless
         of other data members."""
        # TODO: implement for real.
        if self.compat:
            return self.compat[0]

        # Note, we specifically do not hash `_task_options_base` since they are
        # not allowed to impact the results of computation.
        if self._task_options_override:
            task_options_hash = [get_type_registry().get_hash(self._task_options_override)]
        else:
            task_options_hash = []

        if self._hash_includes:
            # Sort to avoid order dependence on the includes.
            hash_includes_hash = sorted(map(get_type_registry().get_hash, self._hash_includes))
        else:
            hash_includes_hash = []

        if self.version is None:
            if self.source:
                source = self.source
            elif self.func:
                source = get_func_source(self.func)
            else:
                source = ""
            return hash_struct(
                ["Task", self.fullname, "source", source] + hash_includes_hash + task_options_hash
            )
        else:
            return hash_struct(
                ["Task", self.fullname, "version", self.version]
                + hash_includes_hash
                + task_options_hash
            )

    def __getstate__(self) -> dict:
        # Note: We specifically don't serialize several items (e.g., func and task_options_base)
        # that are created for us at module load time. These are extracted from the task
        # registry, instead.
        #
        # This needs to remain minimal. See class-level docs before adding anything to this state.
        return {
            "name": self.name,
            "namespace": self.namespace,
            "version": self.version,
            "hash": self.hash,
            "compat": self.compat,
            "script": self.script,
            # This key name is mismatched to avoid a trivial schema update.
            "task_options": self._task_options_override,
        }

    def __setstate__(self, state) -> None:
        self.name = state["name"]
        self.namespace = state["namespace"]
        self.version = state["version"]
        self.hash = state["hash"]
        self.compat = state.get("compat", [])
        self.script = state.get("script", False)
        # This key name is mismatched to avoid a trivial schema update.
        self._task_options_override = state.get("task_options", {})

        # Set func from TaskRegistry.
        registry = get_task_registry()
        _task = registry.get(self.fullname)
        if _task:
            self.func = _task.func
            self._task_options_base = _task._task_options_base
            self.source = _task.source or get_func_source(self.func)
            self._hash_includes = _task._hash_includes
        else:
            self.func = lambda *args, **kwargs: undefined_task(self.fullname, *args, **kwargs)
            self._task_options_base = {}
            self.source = ""
            self._hash_includes = None

        self._signature = None

    def is_valid(self) -> bool:
        """
        Returns True if the Task Value is still valid (task hash matches registry).

        Tasks are first-class Values in redun. They can be cached and fetched
        in future executions. When fetching a Task from the cache, the cached
        hash might no longer exist in the code base (registered tasks).
        """
        return self.hash == self._calc_hash()

    def get_hash(self, data: Optional[bytes] = None) -> str:
        """
        Returns the Task hash.
        """
        return self.hash

    # Note: we can't parameterize PartialTask to a more specific type at this
    # time, due to the complexity of calculating the remaining parameter signature.
    def partial(
        self: "Task[Callable[..., Result]]", *args, **kwargs
    ) -> "PartialTask[Callable[..., Result], Callable[..., Result]]":
        """
        Partially apply some arguments to the Task.
        """
        return PartialTask(self, args, kwargs)

    @property
    def signature(self) -> inspect.Signature:
        """
        Signature of the function wrapped by the task.
        """
        assert self.func
        if not self._signature:
            self._signature = inspect.signature(self.func)
        return self._signature

    @property
    def load_module(self) -> str:
        load_module = self.get_task_option("load_module")
        if load_module:
            return load_module
        else:
            return self.func.__module__


class SchedulerTask(Task[Func]):
    """
    A Task that executes within the scheduler to allow custom evaluation.
    """

    def _call(self: "SchedulerTask[Callable[..., Result]]", *args: Any, **kwargs: Any) -> Result:
        """
        Returns a lazy Expression of calling the task.
        """
        return cast(Result, SchedulerExpression(self.fullname, args, kwargs))

    __call__ = cast(Func, _call)


class PartialTask(Task[Func], Generic[Func, Func2]):
    """
    A Task with only some arguments partially applied.

    The type of this class is parameterized by `Func` and `Func2`, where
    `Func2` is the type of the original function and `Func` is the type
    of partially applied function. They should match on their return types.
    """

    type_name = "redun.PartialTask"

    def __init__(self, task: Task[Func2], args: tuple, kwargs: dict):
        self.task = task
        self.args = tuple(args)
        self.kwargs = kwargs

        super().__init__(task.func, name=task.name, namespace=task.namespace)

    def __repr__(self) -> str:
        return (
            "PartialTask(fullname={fullname}, hash={hash}, args={args}, kwargs={kwargs})".format(
                fullname=self.fullname,
                hash=self.hash[:8],
                args=repr(self.args),
                kwargs=repr(self.kwargs),
            )
        )

    def _call(
        self: "PartialTask[Callable[..., Result], Callable[..., Result]]",
        *args: Any,
        **kwargs: Any,
    ) -> Result:
        # By calling the original task, we ensure that a normal pre-registered
        # task will be the one in the CallGraph recording.
        return self.task(*self.args, *args, **self.kwargs, **kwargs)

    # Cast the signature to match the wrapped function.
    __call__: Func = cast(Func, _call)

    def __getstate__(self) -> dict:
        """
        Returns state for pickling.
        """
        state = super().__getstate__()
        state.update({"task": self.task.__getstate__(), "args": self.args, "kwargs": self.kwargs})
        return state

    def __setstate__(self, state: dict) -> None:
        """
        Sets state from pickle.
        """
        super().__setstate__(state)
        self.task = Task.__new__(Task)
        self.task.__setstate__(state["task"])
        self.args = state["args"]
        self.kwargs = state["kwargs"]

    def _calc_hash(self) -> str:
        return hash_struct(
            [
                "PartialTask",
                self.task._calc_hash(),
                hash_arguments(get_type_registry(), self.args, self.kwargs),
            ]
        )

    def is_valid(self) -> bool:
        return self.task.is_valid()

    def options(
        self: "PartialTask[Callable[..., Result], Callable[..., Result]]",
        **task_options_update: Any,
    ) -> "Task[Func]":
        """
        Returns a new Task with task_option overrides.
        """
        return cast(
            Task[Func],
            self.task.options(**task_options_update).partial(*self.args, **self.kwargs),
        )

    # Note: we can't parameterize PartialTask to a more specific type at this
    # time, due to the complexity of calculating the remaining parameter signature.
    def partial(
        self: "PartialTask[Callable[..., Result], Callable[..., Result]]", *args, **kwargs
    ) -> "PartialTask[Callable[..., Result], Callable[..., Result]]":
        """
        Partially apply some arguments to the Task.
        """
        # Combine new arguments to previously applied arguments.
        args2 = self.args + args
        kwargs2 = {
            **self.kwargs,
            **kwargs,
        }
        return PartialTask(self.task, args2, kwargs2)


@overload
def task(
    func: Func,
) -> Task[Func]:
    ...


@overload
def task(
    *,
    name: Optional[str] = None,
    namespace: Optional[str] = None,
    version: Optional[str] = None,
    compat: Optional[List[str]] = None,
    script: bool = False,
    hash_includes: Optional[list] = None,
    source: Optional[str] = None,
    **task_options_base: Any,
) -> Callable[[Func], Task[Func]]:
    ...


def task(
    func: Optional[Func] = None,
    *,
    name: Optional[str] = None,
    namespace: Optional[str] = None,
    version: Optional[str] = None,
    compat: Optional[List[str]] = None,
    script: bool = False,
    hash_includes: Optional[list] = None,
    source: Optional[str] = None,
    **task_options_base: Any,
) -> Union[Task[Func], Callable[[Func], Task[Func]]]:
    """
    Decorator to register a function as a redun :class:`Task`.

    Parameters
    ----------
    func : Optional[Func]
        A python function to register as a redun Task. If not given, a
        parameterized decorator is returned.
    name : Optional[str]
        Name of task (Default: infer from function `func.__name__`)
    namespace : Optional[str]
        Namespace of task (Default: Infer from context if possible, else None. See
        `Task._compute_namespace`)
    version : Optional[str]
        Optional manual versioning for a task (Default: source code of task is
        hashed).
    compat : Optional[List[str]]
        Optional redun version compatibility. Not currently implemented.
    script : bool
        If True, this is a script-style task which returns a shell script string.
    hash_includes : Optional[list]
        If provided, extra data that should be hashed. That is, extra data that should be
        considered as part of cache invalidation. This list may be reordered without impacting
        the computation. Each list item must be hashable by `redun.value.TypeRegistry.get_hash`.
    source : Optional[str]
        If provided, task.source will be set to this string. It is the caller's responsibility
        to ensure that `source` matches the provided `func` for proper hashing.
    **task_options_base : Any
        Additional options for configuring a task or specifying behaviors of tasks. Since
        these are provided at task construction time (this is typically at Python module-load
        time), they are the "base" set. Example keys:
            load_module : Optional[str]
                The module to load to import this task. (Default: infer from `func.__module__`)
            wrapped_task : Optional[Task]
                If present, a reference to the task wrapped by this one.
    """

    def deco(func: Func) -> Task[Func]:
        nonlocal namespace

        _task: Task[Func] = Task(
            func,
            name=name,
            namespace=namespace,
            version=version,
            compat=compat,
            script=script,
            task_options_base=task_options_base,
            hash_includes=hash_includes,
            source=source,
        )
        get_task_registry().add(_task)
        return _task

    if func:
        # If this decorator is applied directly to a function, decorate it.
        return deco(func)
    else:
        # If a function is not given, just return the parameterized decorator.
        return deco


def scheduler_task(
    name: Optional[str] = None, namespace: Optional[str] = None, version: str = "1"
) -> Callable[[Callable[..., Promise[Result]]], SchedulerTask[Callable[..., Result]]]:
    """
    Decorator to register a function as a scheduler task.

    Unlike usual tasks, scheduler tasks are lower-level tasks that are evaluated
    within the :class:`Scheduler` and allow defining custom evaluation semantics.
    For example, one can implement `cond()`, `seq()` and `catch()` using
    scheduler tasks.

    When evaluated, scheduler tasks are called with a reference to the
    :class:`Scheduler`, the parent :class:`Job`, and the full :class:`SchedulerExpression`
    as it's first three arguments.
    It's remaining arguments are the same as those passed from the user, however,
    they are not evaluated and may contain :class:`Expression`s. It is the
    responsibility of the scheduler task to explicitly evaluate arguments
    by using `Scheduler.evaluate()` as needed. Overall, this allows the scheduler
    task to implement custom evaluation semantics. Lastly, the scheduler task
    must return a :class:`Promise` that resolves to the result of the task.

    This concept corresponds to fexpr in Lisp: https://en.wikipedia.org/wiki/Fexpr

    For example, one could implement a lazy if-statement called `cond` using
    this scheduler task:

    .. code-block:: python

        @scheduler_task()
        def cond(scheduler: Scheduler, parent_job: Job, scheduler_expr: SchedulerExpression,
                 pred_expr: Any, then_expr: Any, else_expr: Any) -> Promise:
            def then(pred):
                if pred:
                    return scheduler.evaluate(then_expr, parent_job=parent_job)
                else:
                    return scheduler.evaluate(else_expr, parent_job=parent_job)

            return scheduler.evaluate(pred_expr, parent_job=parent_job).then(then)

    Once defined, the new `cond` expression can be used like this:

    .. code-block:: python

        @task()
        def main():
            result = task1()
            return cond(result, task2(), task3())
    """

    def deco(func: Callable[..., Promise[Result]]) -> SchedulerTask[Callable[..., Result]]:
        nonlocal name, namespace

        _task: SchedulerTask[Callable[..., Result]] = SchedulerTask(
            func, name=name, namespace=namespace, version=version
        )
        get_task_registry().add(_task)
        return _task

    return deco


def wraps_task(
    wrapper_name: Optional[str] = None,
    wrapper_hash_includes: list = [],
    **wrapper_task_options_base: Any,
) -> Callable[[Callable[[Task[Func]], Func]], Callable[[Func], Task[Func]]]:
    """A helper for creating new decorators that can be used like `@task`, that allow us to
    wrap the task in another one. Conceptually inspired by `@functools.wraps`, which makes it
    easier to create decorators that enclose functions.

    Specifically, this helps us create a decorator that accepts a task and wraps it. The task
    passed in is moved into an inner namespace, hiding it. Then a new task is created from the
    wrapper implementation that assumes its identity; the wrapper is given access to both the
    run-time arguments and the hidden task definition. Since `Tasks` may be wrapped repeatedly,
    the hiding step is recursive. The `load_module` for all of the layers is dictated by the
    innermost concrete `Task` object.

    A simple usage example is a new decorator `@doubled_task`, which simply runs the original
    task and multiplies by two.

    .. code-block:: python

        def doubled_task() -> Callable[[Func], Task[Func]]:

            # The name of this inner function is used to create the nested namespace,
            # so idiomatically, use the same name as the decorator with a leading underscore.
            @wraps_task()
            def _doubled_task(inner_task: Task) -> Callable[[Task[Func]], Func]:

                # The behavior when the task is run. Note we have both the task and the
                # runtime args.
                def do_doubling(*task_args, **task_kwargs) -> Any:
                    return 2 * inner_task.func(*task_args, **task_kwargs)

                return do_doubling
            return _doubled_task

        # The simplest use is to wrap a task that is already created
        @doubled_task()
        @task()
        def value_task1(x: int):
            return 1 + x

        # We can skip the inner decorator and the task will get created implicitly
        # Use the explicit form if you need to pass arguments to task creation.
        @doubled_task()
        def value_task2(x: int):
            return 1 + x

        # We can keep going
        @doubled_task()
        @doubled_task()
        def value_task3(x: int):
            return 1 + x

    There is an additional subtlety if the wrapper itself accepts arguments. These must be passed
    along to the wrapper so they are visible to the scheduler. Needing to do this manually is
    the cost of the extra powers we have.

    .. code-block:: python

        # An example of arguments consumed by the wrapper
        def wrapper_with_args(wrapper_arg: int) -> Callable[[Func], Task[Func]]:

            # WARNING: Be sure to pass the extra data for hashing so it participates in the cache
            # evaluation
            @wraps_task(wrapper_hash_includes=[wrapper_arg])
            def _wrapper_with_args(inner_task: Task) -> Callable[[Task[Func]], Func]:

                def do_wrapper_with_args(*task_args, **task_kwargs) -> Any:
                    return wrapper_arg * inner_task.func(*task_args, **task_kwargs)

                return do_wrapper_with_args
            return _wrapper_with_args

    Parameters
    ----------
    wrapper_name : Optional[str]
        The name of the wrapper, which is used to create the inner namespace. (Default: infer
        from the wrapper `wrapper_func.__name__`)
    wrapper_task_options_base : Any
        Additional options for the wrapper task.
    wrapper_hash_includes : Optional[list]
        If provided, extra data that should be hashed. That is, extra data that should be
        considered as part of cache invalidation. This list may be reordered without impacting
        the computation. Each list item must be hashable by `redun.value.TypeRegistry.get_hash`
    """

    def transform_wrapper(wrapper_func: Callable[[Task], Func]) -> Callable[[Func], Task[Func]]:
        def create_tasks(inner_func_or_task: Union[Func, Task[Func]]) -> Task[Func]:

            # As a convenience, create the lowest level Task on the fly.
            if isinstance(inner_func_or_task, Task):
                hidden_inner_task = inner_func_or_task
            else:
                hidden_inner_task = task()(inner_func_or_task)

            visible_name = hidden_inner_task.name
            visible_namespace = hidden_inner_task.namespace

            def recursive_rename(task_: Task, suffix: str) -> str:
                # Recurse first, so we rename the innermost first, making room for renames
                # higher up the chain.
                if task_.get_task_option("wrapped_task", None) is not None:
                    task_._task_options_base["wrapped_task"] = recursive_rename(
                        get_task_registry().get(task_name=task_.get_task_option("wrapped_task")),
                        suffix,
                    )

                old_fullname = task_.fullname
                if task_.namespace != "":
                    task_.namespace = f"{task_.namespace}.{suffix}"
                else:
                    task_.namespace = suffix
                # Fix the hash, which we have to do manually.
                task_.recompute_hash()

                new_fullname = task_.fullname
                get_task_registry().rename(old_fullname, new_fullname)
                return new_fullname

            nonlocal wrapper_name
            if not wrapper_name:
                wrapper_name = wrapper_func.__name__

            # *Before* we create the new task, hide the old one
            recursive_rename(hidden_inner_task, wrapper_name)

            # Gather data to propagate the implementation of our children, so that cache
            # invalidation propagates. We don't know how wrapper functions will use the tasks,
            # so take the conservative approach and assume that any change invalidates.
            # For example, may call `func` directly, which would put a whole subtree of evaluation
            # out of view of the scheduler.
            #
            # Technically we don't need the name, since the hash has that already, but it's a lot
            # easier to understand this way.
            wrapped_hash_data = [hidden_inner_task]

            # We're definitely getting a task back
            wrapped_task: Task[Func] = task(
                name=visible_name,
                namespace=visible_namespace,
                wrapped_task=hidden_inner_task.fullname,
                load_module=hidden_inner_task.load_module,
                hash_includes=wrapper_hash_includes + wrapped_hash_data,
                **wrapper_task_options_base,
            )(wrapper_func(hidden_inner_task))

            return wrapped_task

        return create_tasks

    return transform_wrapper


class TaskRegistry:
    """
    A registry of currently registered Tasks.

    The @task() decorator registers tasks to the current registry.
    """

    def __init__(self):
        self._tasks: Dict[str, Task] = {}

    def add(self, task: Task) -> None:
        self._tasks[task.fullname] = task

    def get(self, task_name: Optional[str] = None, hash: Optional[str] = None) -> Optional[Task]:
        if task_name:
            return self._tasks.get(task_name)
        elif hash:
            for task in self._tasks.values():
                if task.hash == hash:
                    return task
            return None
        else:
            raise ValueError("No task field given.")

    def rename(self, old_name: str, new_name: str) -> None:
        assert old_name in self._tasks
        task = self._tasks.pop(old_name)
        self._tasks[new_name] = task

    def __iter__(self) -> Iterable[Task]:
        return iter(self._tasks.values())


# Global singleton task registry.
_task_registry = TaskRegistry()



from sqlalchemy.orm import Session

from redun import Scheduler, task
from redun.backends.db import Argument, CallNode, Execution


def test_record_args():
    @task()
    def task1(a, b):
        return a + b

    @task()
    def task2(x):
        return x

    @task()
    def main():
        return task1(task2(2), task2(3))

    scheduler = Scheduler()
    scheduler.run(main())

    args = scheduler.backend.session.query(Argument).all()

    assert {arg.value_parsed for arg in args} == {2, 3}
    assert args[0].call_node.task.name == "task2"

    args = (
        scheduler.backend.session.query(Argument)
        .join(CallNode, CallNode.call_hash == Argument.call_hash)
        .filter(CallNode.task_hash == task1.hash)
        .order_by(Argument.arg_position)
        .all()
    )

    assert args[0].call_node.task.name == "task1"
    assert args[0].arg_position == 0
    assert args[0].value_parsed == 2
    assert args[0].upstream[0].task.name == "task2"

    assert args[1].call_node.task.name == "task1"
    assert args[1].arg_position == 1
    assert args[1].value_parsed == 3
    assert args[1].upstream[0].task.name == "task2"


def test_record_kwargs():
    @task()
    def task1(a=2, b=3):
        return a + b

    @task()
    def task2(x):
        return x

    @task()
    def main():
        return task1(a=task2(2), b=task2(3))

    scheduler = Scheduler()
    scheduler.run(main())

    args = (
        scheduler.backend.session.query(Argument)
        .join(CallNode, CallNode.call_hash == Argument.call_hash)
        .filter(CallNode.task_hash == task1.hash)
        .order_by(Argument.arg_position)
        .all()
    )

    assert args[0].call_node.task.name == "task1"
    assert args[0].arg_position is None
    assert args[0].arg_key == "a"
    assert args[0].value_parsed == 2
    assert args[0].upstream[0].task.name == "task2"

    assert args[1].call_node.task.name == "task1"
    assert args[1].arg_position is None
    assert args[1].arg_key == "b"
    assert args[1].value_parsed == 3
    assert args[1].upstream[0].task.name == "task2"


def test_record_args_nested():
    @task()
    def task1(a, b):
        return [i + j for i, j in zip(a, b)]

    @task()
    def task2(x):
        return x

    @task()
    def main():
        return task1(
            [task2(2), task2(5)],
            [task2(3), task2(4)],
        )

    scheduler = Scheduler()
    scheduler.run(main())

    args = (
        scheduler.backend.session.query(Argument)
        .join(CallNode, CallNode.call_hash == Argument.call_hash)
        .filter(CallNode.task_hash == task1.hash)
        .order_by(Argument.arg_position)
        .all()
    )

    assert args[0].upstream[0].task.name == "task2"
    assert args[0].upstream[1].task.name == "task2"
    assert args[1].upstream[0].task.name == "task2"
    assert args[1].upstream[1].task.name == "task2"


def test_record_args_dup_upstream(scheduler: Scheduler, session: Session) -> None:
    """
    Ensure an argument could have multiple subvalues from the same upstream.
    """

    @task()
    def task1(x):
        return x

    @task()
    def task2():
        return 1

    @task()
    def main():
        arg = task2()
        return task1([arg, arg])

    assert scheduler.run(main()) == [1, 1]

    execution = session.query(Execution).one()
    [task1_call_node] = [
        call_node
        for call_node in execution.job.call_node.children
        if call_node.task_name == "task1"
    ]
    assert task1_call_node.arguments[0].upstream[0].task_name == "task2"



import os
from typing import cast

import boto3

from redun import task
from redun.backends.db import MAX_VALUE_SIZE_PREVIEW, LargeValue, RedunBackendDb, Value
from redun.config import Config
from redun.file import Dir
from redun.scheduler import Scheduler
from redun.tests.utils import mock_s3


def test_value_store(tmpdir) -> None:
    """
    Values are stored on filesystem when configured with value_store_path.
    """

    @task()
    def add1(x: int) -> int:
        return x + 1

    value_store_path = os.path.join(tmpdir, "values/")

    # Use the default value store min size.
    config = {"backend": {"value_store_path": value_store_path}}
    scheduler = Scheduler(config=Config(config))
    assert scheduler.run(add1(10)) == 11

    # All of these values are below the min size so we do not expect anything
    # written in the value store.
    directory = Dir(value_store_path)
    assert not directory.exists()

    # Generate new values.
    @task()
    def add100(x: int) -> int:
        return x + 100

    # Set the value store min size to zero.
    config = {"backend": {"value_store_path": value_store_path, "value_store_min_size": "0"}}
    scheduler = Scheduler(config=Config(config))
    assert scheduler.run(add100(11)) == 111

    # Expect all values written to the value store.
    directory = Dir(value_store_path)
    assert directory.exists()
    assert len(list(directory)) == 3  # Input, output and task.


@mock_s3
def test_value_store_s3() -> None:
    """
    Values are stored on s3 when configured with proper value_store_path.
    """

    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")

    @task()
    def add1(x: int) -> int:
        return x + 1

    value_store_path = "s3://example-bucket/values/"

    # Use the default value store min size.
    config = {"backend": {"value_store_path": value_store_path}}
    scheduler = Scheduler(config=Config(config))
    assert scheduler.run(add1(10)) == 11

    # All of these values are below the min size so we do not expect anything
    # written in the value store.
    directory = Dir(value_store_path)
    assert not directory.exists()

    # Generate new values.
    @task()
    def add100(x: int) -> int:
        return x + 100

    # Set the value store min size to zero.
    config = {"backend": {"value_store_path": value_store_path, "value_store_min_size": "0"}}
    scheduler = Scheduler(config=Config(config))
    assert scheduler.run(add100(11)) == 111

    # Expect all values written to the value store.
    directory = Dir(value_store_path)
    assert directory.exists()
    assert len(list(directory)) == 3  # Input, output and task.


def test_preview_value(tmpdir) -> None:
    """
    We should be able to preview large values.
    """

    @task()
    def make_str(n: int) -> str:
        return "A" * n

    # Setup a scheduler with a value store.
    value_store_path = os.path.join(tmpdir, "values/")
    config = {"backend": {"value_store_path": value_store_path, "value_store_min_size": "100"}}
    scheduler = Scheduler(config=Config(config))
    backend = cast(RedunBackendDb, scheduler.backend)
    assert backend.session

    # Let's make a small value.
    small_result = scheduler.run(make_str(3))
    assert small_result == "AAA"

    # Let's make a large value.
    value_size = MAX_VALUE_SIZE_PREVIEW + 10
    large_result = scheduler.run(make_str(value_size))
    assert len(large_result) == value_size

    # Small value should preview as itself.
    value_hash = scheduler.type_registry.get_hash(small_result)
    value_row = backend.session.query(Value).filter(Value.value_hash == value_hash).one()
    assert value_row.preview == small_result

    # Large value should preview as a LargeValue object.
    value_hash = scheduler.type_registry.get_hash(large_result)
    value_row = backend.session.query(Value).filter(Value.value_hash == value_hash).one()
    preview = value_row.preview
    assert isinstance(preview, LargeValue)
    assert str(preview) == "builtins.str(hash=c901f470, size=1000020)"
